"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import os
import logging
import datetime

from ..version import __version__

logging.basicConfig(level=logging.INFO, format='%(message)s', force=True)
_size_log = 79


# ========================================================================== #
# ========================================================================== #
class _Log:
    """

    """
    def __init__(self, logfile, restart=False):
        if not restart:
            if os.path.isfile(logfile):
                prev_step = 1
                while os.path.isfile(logfile + '{:04d}'.format(prev_step)):
                    prev_step += 1
                os.rename(logfile, logfile + "{:04d}".format(prev_step))

        self.logger_log = logging.getLogger(__name__)
        self.logger_log.addHandler(logging.FileHandler(logfile, 'a'))
        self.logger_log.setLevel(logging.INFO)

        if not restart:
            self.write_header()
        else:
            self.write_restart()

# ========================================================================== #
    def write(self, msg="", center=False, underline=False):
        """
        """
        if underline:
            ul = "*" * len(msg)
        if center:
            msg = msg.center(_size_log, " ").rstrip()
            if underline:
                ul = ul.center(_size_log, " ").rstrip()
        if underline:
            msg = msg + "\n" + ul
        self.logger_log.info(msg)

# ========================================================================== #
    def write_header(self):
        self._delimiter()
        self.write_self()
        self.write()
        self.write_copyright()
        self.write("Please read ACKNOWLEDGMENTS.md for suggested", True)
        self.write("acknowledgments of the MLACS effort.", True)
        self._delimiter()
        self.write()
        self.write(f"version {__version__}")
        self.write()
        now = datetime.datetime.now()
        self.write(f"date: {now.strftime('%d-%m-%Y %H:%M:%S')}")
        self.write()
        self.write()

# ========================================================================== #
    def write_footer(self):
        self._delimiter()
        self.write_copyright()
        self._delimiter()
        self.write()
        self._delimiter()
        self.write("Suggested acknowledgments of the MLACS usage", True, True)
        self.write()
        self.citations()
        self.write("The MLACS package")
        msg = "A. Castellano, R. Béjaud, P. Richard, O. Nadeau, " + \
              "G. Geneste, \nG. Antonius, J. Bouchet, A. Levitt, G. Stoltz" + \
              ", F. Bottin\n" + \
              "(To be submitted (2024))"
        self.write(msg)
        self._delimiter()

# ========================================================================== #
    def write_copyright(self):
        self.write("Copyright (C) 2022-2024 MLACS group.", True)
        self.write("MLACS comes with ABSOLUTELY NO WARRANTY.", True)
        self.write("This package is distributed under the terms of the", True)
        self.write("GNU General Public License, see LICENSE.md", True)
        self.write("or http://www.gnu.org.copyleft/gpl.txt.", True)
        self.write()
        self.write("MLACS is common project of the CEA,", True)
        self.write("Université de Liège, Université du Québec à Trois-Rivières",  # noqa
                   True)
        self.write("and other collaborators, see CONTRIBUTORS.md.", True)

# ========================================================================== #
    def _delimiter(self):
        """
        Just a function to delimitate iterations in the log
        """
        self.write("=" * _size_log)


# ========================================================================== #
# ========================================================================== #
class MlacsLog(_Log):
    '''
    Logging class
    '''
    def __init__(self, logfile, restart=False):
        _Log.__init__(self, logfile, restart)

# ========================================================================== #
    def write_self(self):
        """
        """
        self.write("On-the-fly Machine-Learning Assisted Canonical Sampling",
                   True, True)

# ========================================================================== #
    def citations(self):
        """
        """
        self.write("The MLACS theory and algorithm")
        msg = "A. Castellano, F. Bottin, J. Bouchet, A. Levitt, G. Stoltz" + \
              "\nPhys. Rev. B 106, L161110 (2022)"
        self.write(msg)
        self.write()

# ========================================================================== #
    def write_restart(self):
        self.write()
        self._delimiter()
        self.write("Restarting simulation", True)
        self._delimiter()
        now = datetime.datetime.now()
        self.write(f"date: {now.strftime('%d-%m-%Y %H:%M:%S')}")
        self.write()
        self.write()

# ========================================================================== #
    def write_end(self, isearlystop=False):
        """

        """
        self.write()
        self._delimiter()
        if isearlystop:
            self.write("Convergence criteria reached, stopping the simulation",
                       True)
        else:
            self.write("Max number of step reached, stopping the simulation",
                       True)
        self._delimiter()
        self.write()

# ========================================================================== #
    def write_copyright(self):
        self.write("Copyright (C) 2022-2024 MLACS group.", True)
        self.write("MLACS comes with ABSOLUTELY NO WARRANTY.", True)
        self.write("This package is distributed under the terms of the", True)
        self.write("GNU General Public License, see LICENSE.md", True)
        self.write("or http://www.gnu.org.copyleft/gpl.txt.", True)
        self.write()
        self.write("MLACS is a common project of the CEA,", True)
        self.write("Université de Liège, Université du Québec à Trois-Rivières",  # noqa
                   True)
        self.write("and other collaborators, see CONTRIBUTORS.md.", True)

# ========================================================================== #
    def recap_mlip(self, mlip_params):
        rcut = mlip_params['rcut']
        model = mlip_params['model']
        msg = "Machine-Learning Interatomic Potential parameters\n"
        msg += f"The model used is {model}\n"
        msg += f"Cutoff radius:                               {rcut}\n"
        msg += "Descriptor\n"
        if mlip_params["style"] == "snap":
            twojmax = mlip_params['parameters']['twojmax']
            msg += "Spectral Neighbor Analysis Potential\n"
            msg += f"2Jmax:                                       {twojmax}\n"
            if mlip_params['parameters']["chemflag"] == 1:
                msg += "Multi-element version\n"
        elif mlip_params["style"] == "so3":
            nmax = mlip_params['parameters']['nmax']
            lmax = mlip_params['parameters']['lmax']
            alpha = mlip_params['parameters']['alpha']
            msg += "Smooth SO(3) Power Spectrum\n"
            msg += f"nmax                                         {nmax}\n"
            msg += f"lmax                                         {lmax}\n"
            msg += f"alpha                                        {alpha}\n"
        if mlip_params['regularization'] is not None:
            lam = mlip_params['regularization']
            msg += f"L2-norm regularization with lambda           {lam}\n"
        ecoef = mlip_params['energy_coefficient']
        fcoef = mlip_params['forces_coefficient']
        scoef = mlip_params['stress_coefficient']
        msg += f"Energy coefficient                           {ecoef}\n"
        msg += f"Forces coefficient                           {fcoef}\n"
        msg += f"Stress coefficient                           {scoef}\n"
        msg += "\n"
        self.logger_log.info(msg)

# ========================================================================== #
    def write_input_atoms(self, atoms):
        """
        """
        pos = atoms.get_scaled_positions(wrap=False)
        cell = atoms.get_cell()

        msg = "Initial configuration:\n"
        msg += "----------------------\n"
        msg += "Number of atoms:         {:}\n".format(len(atoms))
        msg += "Elements:\n"
        i = 0
        for symb in atoms.get_chemical_symbols():
            msg += symb + '  '
            i += 1
            if i == 10:
                i = 0
                msg += '\n'
        msg += "\n"
        msg += "\n"
        msg += "Supercell vectors in angstrom:\n"
        for alpha in range(3):
            a, b, c = cell[alpha]
            msg += f'{a:18.16f}  {b:18.16f}  {c:18.16f}\n'
        msg += '\n'
        msg += 'Reduced positions:\n'
        for iat in range(len(atoms)):
            x, y, z = pos[iat]
            msg += '{x:16.16f}  {y:20.16f}  {z:20.16f}\n'
        msg += '\n'
        msg += '\n'
        msg += '============================================================\n'
        self.logger_log.info(msg)

# ========================================================================== #
    def init_new_step(self, step):
        self._delimiter()
        self.write(f"Step {step}")


# ========================================================================== #
class ThermoLog(_Log):
    """
    Logging class for the thermodynamic integration module
    """
    def __init__(self, logfile: str, restart=False):

        if not restart:
            if os.path.isfile(logfile):
                prev_step = 1
                while os.path.isfile(logfile + '{:04d}'.format(prev_step)):
                    prev_step += 1
                os.rename(logfile, logfile + "{:04d}".format(prev_step))

        self.logger_log = logging.getLogger('output')
        self.logger_log.addHandler(logging.FileHandler(logfile, 'a'))

        if not restart:
            self.write_header()
        else:
            self.write_restart()

# ========================================================================== #
    def write_self(self):
        self.write("Non equilibrium thermodynamic integration",
                   True, True)

# ========================================================================== #
    def citations(self):
        """
        """
        self.write("Non-equilibrium thermodynamic integration for solids")
        msg = "R. Freitas, M. Asta and M. de Koning"
        msg += "\nComput Mater. Sci. 112, 333 (2016)"
        self.write(msg)
        self.write()
        self.write("Non-equilibrium thermodynamic integration for fluids")
        msg = "R.P. Leite and M. de Koning"
        msg += "\nComput Mater. Sci. 159, 316 (2018)"
        self.write(msg)
        self.write()
        self.write("MLACS implementation")
        msg = "P Richard, A Castellano, R Béjaud, L Baguet, J Bouchet, "
        msg += "G Geneste, F Bottin"
        msg += "\nPhys. Rev. Lett. 131, 201101 (2023)"
        self.write(msg)
        self.write()



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ase.io import write

from .state import StateManager
from ..core import PathAtoms
from ..core.manager import Manager
from ..mlip.calculator import MlipCalculator
from ..utilities import get_elements_Z_and_masses

default_parameters = {}


# ========================================================================== #
# ========================================================================== #
class BaseMepState(StateManager):
    """
    Class to manage Minimum Energy Path sampling with ASE Optimizers.

    Parameters
    ----------
    images: :class:`list` or `PathAtoms`
        mlacs.PathAtoms or list of ase.Atoms object.
        The list contain initial and final configurations of the reaction path.

    xi: :class:`numpy.array` or `float`
        Value of the reaction coordinate for the constrained MD.
        Default ``None``

    nimages : :class:`int` (optional)
        Number of images used along the reaction coordinate. Default ``1``.
        which is suposed the saddle point.

    mode: :class:`float` or :class:`string`
        Value of the reaction coordinate or sampling mode:
        - ``float`` sampling at a precise coordinate.
        - ``rdm_true`` randomly return the coordinate of an images.
        - ``rdm_spl`` randomly return the coordinate of a splined images.
        - ``rdm_memory`` homogeneously sample the splined reaction coordinate.
        - ``None`` return the saddle point.
        Default ``saddle``

    model: :class:`LinearPotential` or :class:`DeltaLearningPotential`
        mlacs.mlip linear object.
        Default ``None``

    optimizer: :class:`ase.optimize`
        Optimizer from ase.optimize.
        Default :class:`BFGS`

    ftol: :class:`float`
        Stopping tolerance for energy
        Default ``5.0e-2``

    interpolate: :class:`str`
        Method for position interpolation,
        linear or idpp (Image dependent pair potential).
        Default ``linear``

    parameters: :class:`dict` (optional)
        Parameters for ase.neb.NEB class.

    """
    def __init__(self, images, xi=None, nimages=4, mode=None, model=None,
                 interpolate='linear', parameters={}, print=True, **kwargs):

        super().__init__(**kwargs)

        self.model = model
        self.interpolate = interpolate
        self.parameters = default_parameters
        self.parameters.update(parameters)
        self.print = print
        self.patoms = images
        self.nreplica = nimages
        if not isinstance(self.patoms, PathAtoms):
            self.patoms = PathAtoms(images)
            img = [self.patoms.initial]
            img += [self.patoms.initial.copy() for i in range(self.nreplica)]
            img += [self.patoms.final]
            self.patoms.images = img
        if self.model is None:
            raise TypeError('No Calculator defined !')
        if xi is not None:
            self.patoms.xi = xi
        if mode is not None:
            self.patoms.mode = mode

        self.ispimd = False
        self.isrestart = False

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def run_dynamics(self,
                     supercell,
                     pair_style,
                     pair_coeff,
                     model_post=None,
                     atom_style="atomic",
                     eq=False,
                     elements=None):
        """
        Run state function.
        """
        atoms = supercell.copy()
        el, _, _, _ = get_elements_Z_and_masses(atoms)
        if elements is not None and el != elements:
            # ON : Look at LammpsState run_dynamics to implement
            # It is a question of having an MLIP fitted on different elements
            raise NotImplementedError("Need to implement vartypat here")
        initial_charges = atoms.get_initial_charges()

        images = self.patoms.images

        images = self.run_optimize(images)

        self.patoms.images = images
        atoms = self._get_atoms_results(initial_charges)
        return atoms.copy()

# ========================================================================== #
    def run_optimize(self, images):
        """
        Interpolate images and run the optimization.
        """
        pass

# ========================================================================== #
    def _set_calculator(self, images):
        """
        Set Calculator for Forces evaluation.
        """
        pass

# ========================================================================== #
    def _get_atoms_results(self, initial_charges):
        """
        """
        self.patoms.update
        atoms = self.patoms.splined
        if initial_charges is not None:
            atoms.set_initial_charges(initial_charges)
        if self.print:
            write(str(self.folder / 'pos_neb_images.xyz'),
                  self.patoms.images, format='extxyz')
            write(str(self.folder / 'pos_neb_splined.xyz'),
                  self.patoms.splined, format='extxyz')
        return atoms

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        pass


# ========================================================================== #
# ========================================================================== #
class LinearInterpolation(BaseMepState):
    """
    Class to do a simple Linear interpolation of positions with ASE.
    Can be used with the Image dependent pair potential method.
    """
    def __init__(self, images, xi=None, nimages=4, mode=None,
                 model=None, interpolate='linear',
                 parameters={}, print=False, **kwargs):

        super().__init__(images, xi, nimages, mode, model, interpolate,
                         parameters, print, **kwargs)

# ========================================================================== #
    def run_optimize(self, images):
        """
        Interpolate images and run the optimization.
        """

        # RB in future ASE version NEB should be imported from ase.mep
        from ase.neb import NEB
        neb = NEB(images, **self.parameters)

        if self.interpolate == 'idpp':
            neb.interpolate(method='idpp')
        else:
            neb.interpolate()

        images = self._set_calculator(images)

        return images

# ========================================================================== #
    def _set_calculator(self, images):
        """
        Interpolate images and run the optimization.
        """
        for img in images:
            img.calc = MlipCalculator(self.model)
        return images

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "Linear interpolation\n"
        msg += f"Number of replicas:                 {self.patoms.nreplica}\n"
        msg += f"Interpolation method:               {self.interpolate}\n"
        msg += f"Sampling mode:                      {self.patoms.mode}\n"
        msg += f"Sampled coordinate:                 {self.patoms.xi}\n"
        msg += "\n"
        return msg


# ========================================================================== #
# ========================================================================== #
class NebAseState(BaseMepState):
    """
    Class to run the Nudged Elastic Band method with ASE Optimizers.
    """
    def __init__(self, images, xi=None, nimages=4, mode=None,
                 model=None, interpolate='linear',
                 Kspring=0.1, optimizer=None, ftol=5.0e-2,
                 parameters={}, print=False, **kwargs):

        super().__init__(images, xi, nimages, mode, model, interpolate,
                         parameters, print, **kwargs)

        self.opt = optimizer
        self.criterions = ftol
        self.Kspring = Kspring
        if self.opt is None:
            from ase.optimize import MDMin
            self.opt = MDMin

# ========================================================================== #
    def run_optimize(self, images):
        """
        Interpolate images and run the optimization.
        """

        images = self._set_calculator(images)

        # RB in future ASE version NEB should be imported from ase.mep
        from ase.neb import NEB
        neb = NEB(images, k=self.Kspring, **self.parameters)

        if self.interpolate == 'idpp':
            neb.interpolate(method='idpp')
        else:
            neb.interpolate()

        opt = self.opt(neb)
        opt.run(fmax=self.criterions, steps=self.nsteps)

        images[0].calc = MlipCalculator(self.model)
        images[-1].calc = MlipCalculator(self.model)

        return images

# ========================================================================== #
    def _set_calculator(self, images):
        """
        Interpolate images and run the optimization.
        """
        for img in images[1:-1]:
            img.calc = MlipCalculator(self.model)
        return images

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "NEB calculation as implemented in ASE\n"
        msg += f"Number of replicas:                 {self.patoms.nreplica}\n"
        msg += f"Interpolation method:               {self.interpolate}\n"
        msg += f"Sampling mode:                      {self.patoms.mode}\n"
        msg += f"Sampled coordinate:                 {self.patoms.xi}\n"
        msg += "\n"
        return msg


# ========================================================================== #
# ========================================================================== #
class CiNebAseState(NebAseState):
    """
    Class to run the Climbing Image Nudged Elastic Band method
    with ASE Optimizers.
    """
    def __init__(self, images, xi=None, nimages=3, mode=None,
                 model=None, interpolate='linear',
                 Kspring=0.1, optimizer=None, ftol=5.0e-2,
                 parameters={}, print=False, **kwargs):

        super().__init__(images, xi, nimages, mode, model, interpolate,
                         Kspring, optimizer, ftol,
                         parameters, print, **kwargs)

        self.parameters.update(dict(climb=True))

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "Ci-NEB calculation as implemented in ASE\n"
        msg += f"Number of replicas:                     {self.nreplica}\n"
        msg += f"Interpolation method:                   {self.interpolate}\n"
        msg += f"Sampling mode:                          {self.patoms.mode}\n"
        msg += f"Sampled coordinate:                     {self.patoms.xi}\n"
        msg += "\n"
        return msg


# ========================================================================== #
# ========================================================================== #
class StringMethodAseState(NebAseState):
    """
    Class to run the String Method with ASE Optimizers.
    """
    def __init__(self, images, xi=None, nimages=4, mode=None,
                 model=None, interpolate='linear',
                 Kspring=0.1, optimizer=None, ftol=5.0e-2,
                 parameters={}, print=False, **kwargs):

        super().__init__(images, xi, nimages, mode, model, interpolate,
                         Kspring, optimizer, ftol,
                         parameters, print, **kwargs)

        self.parameters.update(dict(method='string'))

# ========================================================================== #
    def _set_calculator(self, images):
        """
        Interpolate images and run the optimization.
        """
        for img in images:
            img.calc = MlipCalculator(self.model)
        return images

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "String method calculation as implemented in ASE\n"
        msg += f"Number of replicas:                 {self.patoms.nreplica}\n"
        msg += f"Interpolation method:               {self.interpolate}\n"
        msg += f"Sampling mode:                      {self.patoms.mode}\n"
        msg += f"Sampled coordinate:                 {self.patoms.xi}\n"
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.build import bulk
from unittest.mock import patch

from ... import context  # noqa
from mlacs.mlip.descriptor import Descriptor


@patch.multiple(Descriptor, __abstractmethods__=set())
def create_obj():
    at = bulk("Si")
    desc = Descriptor(at)
    return at, desc


@patch.multiple(Descriptor, __abstractmethods__=set())
def test_elem():
    """
    Here we test that the descriptor has the right elements,
    Z, masses and charges compared to the input atoms
    """
    at, desc = create_obj()
    ref_elements = np.array(["Si"])
    ref_Z = np.array([14])
    ref_masses = np.array([at.get_masses()[0]])
    ref_charges = None

    assert desc.elements == ref_elements
    assert desc.Z == ref_Z
    assert desc.masses == ref_masses
    assert desc.charges == ref_charges

    ref_charges = np.array([1])
    at.set_initial_charges([1, 1])
    desc = Descriptor(at)
    assert desc.charges == ref_charges



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from ..utilities.plots import plot_correlation, init_rcParams


def main(args, parser):

    data = np.loadtxt(args.file)
    if args.datatype not in ["energy", "forces", "stress", None]:
        raise ValueError("The type argument has to be "
                         "energy, forces or stress")
    if args.density and args.weight is not None:
        raise ValueError("density and weights can't be used at the same time")
    rmse = True
    if args.normse:
        rmse = False
    mae = True
    if args.nomae:
        mae = False
    rsquared = True
    if args.nor2:
        rsquared = False
    density = False
    if args.density:
        density = True
    if args.weight is not None:
        weight = np.loadtxt(args.weight)
    else:
        weight = None
    cmap = args.cmap
    size = float(args.size)
    figsize = (float(args.figsize), float(args.figsize))
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    init_rcParams()
    if args.density or args.weight is not None:
        gd = GridSpec(1, 12, fig)
        ax = fig.add_subplot(gd[:, :11])
        axcbar = fig.add_subplot(gd[:, 11:])
    else:
        ax = fig.add_subplot()
        axcbar = None
    plot_correlation(ax,
                     data,
                     datatype=args.datatype,
                     density=density,
                     weight=weight,
                     showrmse=rmse,
                     showmae=mae,
                     showrsquared=rsquared,
                     cmap=cmap,
                     size=size,
                     marker=args.marker,
                     axcbar=axcbar)
    ax.set_aspect("equal")
    if args.save is not None:
        plt.savefig(args.save)
    if not args.noshow:
        plt.show()


class CLICommand:
    """Plot correlation of DFT and MLIP data

    Read data points in the file and plot the correlation along with
    the RMSE, MAE and M^2

    Example:

        $ mlacs correlation MLIP-Energy_comparison.dat --datatype energy
    """
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('file', help="file with the data")
        parser.add_argument("-s", "--save", default=None,
                            help="Name of the file to save the plot. "
                                 "Default None")
        parser.add_argument("--noshow", action="store_true",
                            help="To disable the visualisation of the plot")
        parser.add_argument('--datatype', default=None,
                            help="Type of data in the file. Can be "
                            "energy, forces or stress")
        parser.add_argument('--density', action="store_true",
                            help="Color points according to the density")
        parser.add_argument('--weight', default=None,
                            help="Color points according to MBAR weights")
        parser.add_argument('--nomae', action="store_true",
                            help="To remove the MAE from the plot")
        parser.add_argument('--normse', action="store_true",
                            help="To remove rmse from the plot")
        parser.add_argument('--nor2', action="store_true",
                            help="to remove the r^2 from the plot")
        parser.add_argument("--figsize", default="10",
                            help="Size of the figure for matplotlib")
        parser.add_argument("--cmap", default="inferno",
                            help="Colormap for the density plot")
        parser.add_argument("--size", default=5,
                            help="Size of the marker")
        parser.add_argument("--marker", default='o',
                            help="Markers for the data")

    @staticmethod
    def run(args, parser):
        main(args, parser)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np

from .lammps_state import BaseLammpsState
from ..core.manager import Manager
from ..utilities.io_lammps import (LammpsBlockInput,
                                   write_atoms_lammps_spin_style)


_default_params = dict(temperature=300,
                       t_stop=None,
                       damp=None,
                       dt=1.5,
                       fixcm=True)


# ========================================================================== #
# ========================================================================== #
class SpinLammpsState(BaseLammpsState):
    """
    Class to manage magnetic spin simulations with LAMMPS and the SPIN package.

    Tranchida, et al., Journal of Computational Physics, 372, 406-425, (2018).

    Parameters
    ----------
    temperature: :class:`float`
        Temperature of the simulation, in Kelvin.

    t_stop: :class:`float` or ``None`` (optional)
        When this input is not ``None``, the temperature of
        the molecular dynamics simulations is randomly chosen
        in a range between `temperature` and `t_stop`.
        Default ``None``

    damp: :class:`float` or ``None`` (optional)

    dt : :class:`float` (optional)
        Timestep, in fs. Default ``1.5`` fs.

    nsteps : :class:`int` (optional)
        Number of MLMD steps for production runs. Default ``1000`` steps.

    nsteps_eq : :class:`int` (optional)
        Number of MLMD steps for equilibration runs. Default ``100`` steps.

    fixcm : :class:`Bool` (optional)
        Fix position and momentum center of mass. Default ``True``.

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    rng : RNG object (optional)
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    init_momenta : :class:`numpy.ndarray` (optional)
        Gives the (Nat, 3) shaped momenta array that will be used
        to initialize momenta when using
        the `initialize_momenta` function.
        If the default ``None`` is set, momenta are initialized with a
        Maxwell Boltzmann distribution.
    """
    def __init__(self, temperature, t_stop=None, damp=None, dt=1.5, fixcm=True,
                 nsteps=1000, nsteps_eq=100, logfile=None, trajfile=None,
                 loginterval=50, blocks=None, **kwargs):

        super().__init__(nsteps, nsteps_eq, logfile, trajfile, loginterval,
                         blocks, **kwargs)

        self.temperature = temperature
        self.t_stop = t_stop
        self.damp = damp
        self.dt = dt
        self.fixcm = fixcm

        if self.rng is None:
            self.rng = np.random.default_rng()
        if self.damp is None:
            self.damp = "$(100*dt)"

# ========================================================================== #
    def _get_block_init(self, atom_style, pbc, el, masses):
        """

        """
        pbc = "{0} {1} {2}".format(*tuple("sp"[int(x)] for x in pbc))

        block = LammpsBlockInput("init", "Initialization")
        block("map", "atom_modify map array")
        block("units", "units metal")
        block("boundary", f"boundary {pbc}")
        block("atom_style", "atom_style spin")
        block("read_data", "read_data atoms.in")
        for i, mass in enumerate(masses):
            block(f"mass{i}", f"mass {i+1}  {mass}")
        return block

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """
        Function to write the thermostat of the mlmd run
        """
        if self.t_stop is None:
            temp = self.temperature
        else:
            tmp_temp = np.sort([self.temperature, self.t_stop])
            if eq:
                temp = np.max(tmp_temp)
            else:
                temp = self.rng.uniform(*tmp_temp)
        langevinseed = self.rng.integers(1, 9999999)
        spinseed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("thermostat", "Thermostat")
        block("timestep", f"timestep {self.dt / 1000}")
        block("rmv_langevin", "fix ff all store/force")

        txt = f"fix f1 all langevin {temp} {temp} {self.damp} " + \
              f"{langevinseed} gjf {self.gjf} zero yes"
        block("langevin", txt)
        txt = f"fix fspin all langevin/spin {temp} {self.damp} {spinseed}"
        block("nvt_spin", txt)
        block("nve_spin", "fix fspin2 all nve/spin lattice moving")
        if self.fixcm:
            block("cm", "fix fcm all recenter INIT INIT INIT")
        return block

# ========================================================================== #
    def _get_block_traj(self, el):
        """

        """
        block = LammpsBlockInput("traj", "Dumping trajectory")
        txt = "compute spin all property/atom sp spx spy spz fmx fmy fmz"
        block("compute", txt)
        txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} " + \
              "id type xu yu zu vx vy vz fx fy fz " + \
              "f_ff[1] f_ff[2] f_ff[3] c_spin[1] " + \
              "c_spin[2] c_spin[3] c_spin[4] c_spin[5] c_spin[6] c_spin[7]" + \
              " element"
        block("dump", txt)
        block("dump_modify1", "dump_modify dum1 append yes")
        txt = "dump_modify dum1 element " + " ".join([p for p in el])
        block("dump_modify2", txt)
        return block

# ========================================================================== #
    def _get_block_log(self):
        """

        """
        block = LammpsBlockInput("log", "Logging")
        # First we compute the magnetic quantities
        block("compute", "compute out_mag all spin")
        block("magx", "variable magx equal c_out_mag[1]")
        block("magy", "variable magy equal c_out_mag[2]")
        block("magz", "variable magz equal c_out_mag[3]")
        block("magn", "variable magn equal c_out_mag[4]")
        block("mage", "variable mage equal c_out_mag[5]")
        block("magt", "variable magt equal c_out_mag[6]")

        variables = ["t equal step", "mytemp equal temp",
                     "mype equal pe", "myke equal ke", "myetot equal etotal",
                     "mypress equal press/10000", "vol equal (lx*ly*lz)"]
        for i, var in enumerate(variables):
            block(f"variable{i}", f"variable {var}")
        txt = f"fix mylog all print {self.loginterval} \""
        variables = ["$t", "${mytemp}", "$(v_magt)", "${vol}", "${myetot}",
                     "${myke}", "$(v_mage)", "${mypress}",
                     "$(v_magx)", "$(v_magy)", "$(v_magz)", "$(v_magn)"]
        txt += " ".join(variables) + "\" "
        txt += f"append {self.logfile} title "
        titles = ["Step", "T_at", "T_spin", "Vol", "Etot", "Epot", "Ekin",
                  "Espin", "Press", "Magn_x", "Magn_y", "Magn_z", "Magn_norm"]
        txt += "\"# " + " ".join(titles) + "\""
        block("fix", txt)
        return block

# ========================================================================== #
    def _get_block_lastdump(self, el, eq):
        block = LammpsBlockInput("lastdump", "Dump last configuration")
        txt = "dump last all custom 1 configurations.out " + \
              "id type xu yu zu vx vy vz fx fy fz c_spin[1] " + \
              "c_spin[2] c_spin[3] c_spin[4] c_spin[5] c_spin[6] c_spin[7]" + \
              " element"
        block("dump", txt)
        txt = "dump_modify last element " + " ".join([p for p in el])
        block("dump_modify1", txt)
        block("run_dump", "run 0")
        return block

# ========================================================================== #
    @Manager.exec_from_path
    def _write_lammps_atoms(self, atoms, atom_style, elements=None):
        """

        """
        spins = atoms.get_array("spins")
        with open("atoms.in", "w") as fd:
            write_atoms_lammps_spin_style(fd, atoms, spins)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .pdf import compute_pdf
from .miscellanous import (get_elements_Z_and_masses,
                           create_random_structures,
                           compute_averaged,
                           interpolate_points,
                           compute_correlation,
                           integrate_points,
                           normalized_integration,
                           execute_from,
                           save_cwd,
                           create_link,
                           )

from .io_abinit import (AbinitNC)

from .io_pandas import (make_dataframe)

__all__ = ['compute_pdf',
           'get_elements_Z_and_masses',
           'create_random_structures',
           'compute_averaged',
           'interpolate_points',
           'compute_correlation',
           'integrate_points',
           'normalized_integration',
           'AbinitNC',
           'execute_from',
           'save_cwd',
           'create_link',
           'make_dataframe',
           ]



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
import netCDF4 as nc

from ..core.manager import Manager
from .calc_property import (CalcRoutineFunction,
                            CalcPressure,
                            CalcAcell,
                            CalcAngles,
                            CalcSpinAt,
                            CalcElectronicEntropy)


# ========================================================================== #
# ========================================================================== #
class PropertyManager(Manager):
    """
    Parent Class managing the calculation of differents properties
    """

    def __init__(self,
                 prop,
                 folder='Properties',
                 **kwargs):

        if prop is None:
            self.check = [False]
            self.manager = None
            folder = ''
        elif isinstance(prop, list):
            self.manager = prop
            self.check = [False for _ in range(len(prop))]
        else:
            self.manager = [prop]
            self.check = [False]

        if prop is not None:
            if not any([prop.needdir for prop in self.manager]):
                folder = ''

        Manager.__init__(self, folder=folder, **kwargs)
        if self.manager is not None:
            for observable in self.manager:
                if observable.needdir:
                    observable.workdir = self.workdir
                    observable.folder = self.folder

# ========================================================================== #
    @property
    def check_criterion(self):
        """
        Check all criterions. They have to be all converged at the same time.
        Return True if all elements in list are True, else return False.
        """
        return np.all(self.check)

# ========================================================================== #
    @Manager.exec_from_subdir
    def run(self, step):
        """
        Run property calculation.
        """
        msg = ""
        for i, observable in enumerate(self.manager):
            if step % observable.freq == 0:
                if observable.needdir:
                    observable.workdir = self.workdir
                    observable.folder = self.folder
                    observable.subfolder = f'Step{step}'
                self.check[i] = observable._exec()
                msg += repr(observable)
        return msg

# ========================================================================== #
    def calc_initialize(self, **kwargs):
        """
        Add on the fly arguments for calculation of properties.
        """
        for observable in self.manager:
            if observable.useatoms:
                observable.get_atoms(kwargs['atoms'])

# ========================================================================== #
    def save_prop(self, step):
        """
        Save the values of observables contained in a PropertyManager object.

        Parameters
        ----------

        step: :class:`int`
            The index of MLAS iteration
        """

        if self.manager is not None:
            for observable in self.manager:
                if step % observable.freq == 0:
                    to_be_saved = observable.new
                    nc_name = observable.nc_name

                    if nc_name is not None:
                        ncpath = self.ncfile.ncpath
                        for idx, val_state in enumerate(to_be_saved):
                            with nc.Dataset(ncpath, 'a') as ncfile:
                                index_state = idx+1
                                metadata = [step, index_state]
                                idx_db = np.ma.count(
                                        ncfile[nc_name+'_meta'][:, 0])
                                # idx_db is index of conf in dtbase
                                # for observable
                                ncfile[nc_name][idx_db] = val_state
                                ncfile[nc_name+'_meta'][idx_db] = metadata
                                ncfile['mdtime'][idx_db] = idx_db + 1

# ========================================================================== #
    def save_weighted_prop(self, step, weighting_pol):
        """
        For all observables in a PropertyManager object, save the values
        of the observables, weighted by the weighting policy.

        Parameters
        ----------

        step: :class:`int`
            The index of MLAS iteration

        weighting_pol: :class:`WeightingPolicy`
            WeightingPolicy class.

        """
        ncpath = self.ncfile.ncpath

        if weighting_pol is not None:
            for observable in self.manager:
                nc_name = observable.nc_name
                weights = weighting_pol.weight[2:].copy()
                obs = self.ncfile.read_obs(nc_name)
                observable_values = obs[:len(weights)]

                if len(weights) > 0:
                    w_name = 'weighted_' + nc_name
                    weights /= np.sum(weights)
                    dim_array = np.array([1]*observable_values.ndim)
                    dim_array[0] = -1
                    r_weights = weights.reshape(dim_array)
                    if r_weights.shape[0] == observable_values.shape[0]:
                        weighted_observ = np.sum(r_weights*observable_values)
                        with nc.Dataset(ncpath, 'a') as ncfile:
                            ncfile[w_name][len(weights)-1] = weighted_observ

# ========================================================================== #
    def save_weights(self, step, weighting_pol, ncformat):
        """
        Save the MBAR weights.

        Parameters
        ----------

        step: :class:`int`
            The index of MLAS iteration

        weighting_pol: :class:`WeightingPolicy`
            WeightingPolicy class, Default: `None`.

        ncformat: :class:`str`
            The format of the *HIST.nc file. One of the five flavors of netCDF
            files format available in netCDF4 python package 'NETCDF3_CLASSIC',
            'NETCDF3_64BIT_OFFSET', 'NETCDF3_64BIT_DATA','NETCDF4_CLASSIC',
            'NETCDF4'.
        """

        if weighting_pol is not None:
            # The first two confs of self.mlip.weight.database are never used
            # in the properties computations, so they are throwned out here
            # by the slicing operator [2:]
            weights = weighting_pol.weight[2:].copy()
            if len(weights) > 0:
                nb_effective_conf = np.sum(weights)**2 / np.sum(weights**2)
            else:
                nb_effective_conf = 0
            nb_conf = len(weights)

            # Save weights into HIST file
            weights_ncpath = self.ncfile.ncpath
            if 'NETCDF3' in ncformat:
                weights_ncpath = weights_ncpath.replace('HIST', 'WEIGHTS')
            with nc.Dataset(weights_ncpath, 'a') as ncfile:
                idx_db = np.ma.count(ncfile['weights_meta'][:, 0])
                for idx, value in enumerate(weights):
                    ncfile['weights'][idx_db+idx] = value
                    # weights_meta keeps track of db index for given cycle
                    # and of number of effective configurations
                    metadata = [idx + 1, nb_effective_conf, nb_conf]
                    ncfile['weights_meta'][idx_db+idx] = metadata

            # Weight of first two confs (that are throwned out)
            w_first2 = abs(np.sum(weighting_pol.weight) - np.sum(weights))
            return w_first2


# ========================================================================== #
# ========================================================================== #
class RoutinePropertyManager(PropertyManager):
    """
    Class to handle the list of CalcRoutineFunction.

    Parameters
    ----------

    ncfile: :class:`OtfMlacsHist`
        The netcdf *HIST.nc file.
    """

    def __init__(self, ncfile, launched):

        # Get variables names, dimensions, and units conventions
        var_dim_dict = ncfile.var_dim_dict
        ase_units_dict = ncfile.ase_units_dict
        abinit_units_dict = ncfile.abinit_units_dict

        # Build RoutinePropertyManager
        routine_prop_list = []
        for x in var_dim_dict:
            var_name, var_dim = var_dim_dict[x]
            var_abinit_unit = abinit_units_dict[x]
            var_ase_unit = ase_units_dict[x]
            lammps_func = 'get_' + x.lower()
            observable = CalcRoutineFunction(lammps_func,
                                             label=x,
                                             nc_name=var_name,
                                             nc_dim=var_dim,
                                             nc_unit=var_abinit_unit,
                                             ase_unit=var_ase_unit,
                                             frequence=1)
            routine_prop_list.append(observable)
        other_observables = [CalcPressure(), CalcAcell(), CalcAngles(),
                             CalcSpinAt(), CalcElectronicEntropy()]
        routine_prop_list += other_observables

        PropertyManager.__init__(self, prop=routine_prop_list)

        if not launched:
            ncfile.create_nc_var(self.manager)

        self.workdir = self.workdir
        self.isfirstlaunched = not launched
        self.ncfile = ncfile



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""


from ase.calculators.socketio import SocketIOCalculator

from mlacs.calc.calc_manager import CalcManager


# =========================================================================== #
# =========================================================================== #
class SocketCalcManager(CalcManager):
    """
    Class for managing the true potential through the SocketIO calculator
    Work in progress.

    Parameters
    ----------
    """
    def __init__(self,
                 calc=None,
                 magmoms=None,
                 socketlog=None,
                 unixsocket=None,
                 port=None,
                 **kwargs):
        CalcManager.__init__(self, calc=calc, magmoms=magmoms, **kwargs)

        # This will launch the server
        SocketIOCalculator(unixsocket=unixsocket, port=port, log=socketlog)

# =========================================================================== #
    def compute_true_potential(self, atoms):
        """
        """
        atoms.set_initial_magnetic_moments(self.magmoms)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
import os
import shutil
import numpy as np
import shlex

# IMPORTANT : subprocess->Popen doesnt work if we import run, PIPE
from subprocess import Popen
from subprocess import check_output
import logging
from concurrent.futures import ThreadPoolExecutor

from ase import Atom
from ase.symbols import symbols2numbers
from ase.calculators.singlepoint import SinglePointCalculator as SPCalc
from ase.io.abinit import (write_abinit_in,
                           read_abinit_out)

from ..core.manager import Manager
from .calc_manager import CalcManager
from ..utilities import save_cwd
from ..utilities.io_abinit import AbinitNC


# ========================================================================== #
# ========================================================================== #
class AbinitManager(CalcManager):
    """
    This Calc class is an extended object for Abinit calculators.
    The AbinitManager can handdle netCDF files, MPI processes and a better
    pseudopotentials files management.

    Parameters
    ----------
    parameters: :class:`dict`
        Dictionnary of abinit input

    pseudos: :class:`dict`
        Dictionnary for the pseudopotentials
        {'O': /path/to/pseudo}

    abinit_cmd: :class:`str`
        The command to execute the abinit binary.

    mpi_runner : :class:`str`
        The command to call MPI.
        I assume the number of processor is specified using -n argument
        Default ``mpirun``

    magmoms: :class:`np.ndarray` (optional)
        An array for the initial magnetic moments for each computation
        If ``None``, no initial magnetization. (Non magnetic calculation)
        Default ``None``.

    folder: :class:`str` (optional)
        The root for the directory in which the computation are to be done
        Default 'DFT'

    logfile: :class:`str` (optional)
        The name of the Abinit log file inside the workdir folder
        Default 'abinit.log'

    errfile: :class:`str` (optional)
        The name of the Abinit error file inside the workdir folder
        Default 'abinit.err'

    nproc: :class:`int` (optional)
        Number of processor available for all Abinit.

    nproc_per_task: :class:`int` (optional)
        Number of processor available for all Abinit.
        Default nproc


    Examples
    --------

    >>> from mlacs.calc import AbinitManager
    >>> variables = dict(ixc=-1012, ecut=12, tsmear=0.001, occopt=3, nband=82,
    >>>                  ngkpt=[2, 2, 2], shiftk=[0.5, 0.5, 0.5],
    >>>                  autoparall=1, nsym=1) # Cu, 8 atoms.
    >>> pseudos = {'Cu': "/path/to/pseudo/Cu.LDA_PW-JTH.xml"}
    >>> calc = AbinitManager(parameters=variables, pseudos=pseudos)
    """

    def __init__(self,
                 parameters,
                 pseudos,
                 abinit_cmd="abinit",
                 mpi_runner="mpirun",
                 magmoms=None,
                 folder='DFT',
                 logfile="abinit.log",
                 errfile="abinit.err",
                 nproc=1,
                 nproc_per_task=None,
                 **kwargs):

        CalcManager.__init__(self, "dummy", magmoms,
                             folder=folder, **kwargs)

        self.parameters = parameters
        if 'IXC' in self.parameters.keys():
            self.parameters['ixc'] = self.parameters['IXC']
            del self.parameters['IXC']
        if 'ixc' not in self.parameters.keys():
            msg = 'WARNING AbinitManager:\n'
            msg += 'You should specify an ixc value or ASE will set 7 (LDA) !'
            msg += '\n(Does not apply if using a PAW xml format)'
            logging.warning(msg)

        self._organize_pseudos(pseudos)
        self.abinit_cmd = abinit_cmd
        self.mpi_runner = mpi_runner
        self.nproc = nproc
        if nproc_per_task is None:
            nproc_per_task = self.nproc
        self.nproc_per_task = nproc_per_task

        self.log = logfile
        self.err = errfile
        self.ncfile = AbinitNC()

# ========================================================================== #
    @staticmethod
    def submit_abinit_calc(cmd, logfile, errfile, cdir):
        with open(logfile, 'w') as lfile, \
                open(errfile, 'w') as efile:
            try:
                process = Popen(cmd,
                                cwd=cdir,
                                stderr=efile,
                                stdout=lfile,
                                shell=False)
                process.wait()
            except Exception as e:
                msg = f"This command {' '.join(cmd)}\n"
                msg += f"raised this exception {e}"
                efile.write(msg)

# ========================================================================== #
    @Manager.exec_from_subdir
    def compute_true_potential(self, confs: [Atom],
                               subfolder: [str],
                               step: [int]):
        """
        Compute the energy/forces/stress of given configurations with Abinit.

        Parameters
        ----------
        confs: :class:`list` of :class:`ase.Atoms`
            The input list of atom objects.

        subfolder: :class:`list` of :class:`str` (optional)
            Subfolder in which the properties are saved.

        step: :class:`list` of :class:`int`  (optional)
            The list of configuration indices.

        Returns
        -------
        result_confs: :class:`list` of :class:`ase.Atoms`
            The output list of atom objects, with corresponding
            SinglePointCalculator resulting from true potential calculation.
            See also _read_output() where SinglePointCalculator() is called.
        """
        assert len(confs) == len(subfolder) == len(step)
        nparal = self.nproc // self.nproc_per_task

        # Prepare all calculations
        confs = [at.copy() for at in confs]
        path_prefix_l = []
        for at, sf, istep in zip(confs, subfolder, step):
            # First set the prefix
            self.subfolder = sf
            self.prefix = str(self.subfolder) + '_'
            # Then append a level to subsubdir
            self.subsubdir = self.subsubdir / f"Step{istep}"
            # Save this list for parallel execution
            path_prefix_l.append((self.subsubdir, self.prefix))
            # Initialize objects
            at.set_initial_magnetic_moments(self.magmoms)
            self._write_input(at)

        # Yeah for threading
        # GA: I would move the threading outside of this function
        # because the files naming depends on external objects.
        with save_cwd(), ThreadPoolExecutor(max_workers=nparal) as executor:
            for (path, pref) in path_prefix_l:
                self.subsubdir = path
                self.prefix = pref
                command = self._make_command()
                executor.submit(self.submit_abinit_calc,
                                command,
                                self.get_filepath(self.log),
                                self.get_filepath(self.err),
                                cdir=str(self.subsubdir))
            executor.shutdown(wait=True)

        # Now we can read everything
        results_confs = []
        for ((path, pref), at) in zip(path_prefix_l, confs):
            self.subsubdir = path
            self.prefix = pref
            results_confs.append(self._read_output(at))

        for i in range(len(results_confs)):
            results_confs[i].info = confs[i].info

        # Reset values for good measure...
        self.prefix = ''
        self.subfolder = ''

        # Tada !
        return results_confs

# ========================================================================== #
    def _make_command(self):
        """
        Make the command to call Abinit including MPI :
        """
        abinit_cmd = self.abinit_cmd + " " + self.get_filepath("abinit.abi")
        nproc = self.nproc_per_task

        if nproc > 1:
            mpi_cmd = "{} -n {}".format(self.mpi_runner, nproc)
        else:
            mpi_cmd = ""

        full_cmd = "{} {}".format(mpi_cmd, abinit_cmd)
        full_cmd = shlex.split(full_cmd, posix=(os.name == "posix"))
        return full_cmd

# ========================================================================== #
    @Manager.exec_from_subdir
    def _write_input(self, atoms):
        """
        Write the input for the current atoms
        """
        if self.subsubdir.exists():
            self._remove_previous_run(str(self.subsubdir))
        else:
            self.subsubdir.mkdir(parents=True)

        # First we need to prepare some stuff
        original_pseudos = self.pseudos.copy()
        species = sorted(set(atoms.numbers))
        self._copy_pseudos()

        unique_elements = set(atoms.get_chemical_symbols())
        pseudos = [pseudo for pseudo, el in zip(self.pseudos, self.typat) if
                   el in unique_elements]

        with open(self.get_filepath("abinit.abi"), "w") as fd:
            write_abinit_in(fd,
                            atoms,
                            self.parameters,
                            species,
                            pseudos)
        self.pseudos = original_pseudos

# ========================================================================== #
    def _copy_pseudos(self):
        """
        Create a unique file for each read/write operation
        to prevent netCDF4 error.
        The path and the filename must be unique.
        """
        def _create_copy(source, dest):
            if os.path.exists(dest):
                return
            else:
                shutil.copy(source, dest)
        new_psp = []
        pp_dirpath = self.parameters.get("pp_dirpath")
        pseudos = self.pseudos

        # Create an unique psp file in the DFT/State/Step folder
        if pp_dirpath is None:
            pp_dirpath = ""
        if isinstance(pseudos, str):
            pseudos = [pseudos]
        for psp in pseudos:
            fn = psp.split('/')[-1]
            source = pp_dirpath+psp
            dest = self.get_filepath(fn)
            _create_copy(source, dest)
            new_psp.append(dest)
        self.pseudos = new_psp
        self.parameters.pop('pspdir', None)

# ========================================================================== #
    def _read_output(self, at):
        """
        """
        results = {}
        if 'None' not in str(self.ncfile):
            ncpath = self.get_filepath("abinito_GSR.nc")
            self.ncfile.read(filename=ncpath)
            atoms = self.ncfile.convert_to_atoms()[0]
            atoms.set_velocities(at.get_velocities())
            return atoms

        with open(self.get_filepath("abinit.abo")) as fd:
            dct = read_abinit_out(fd)
            results.update(dct)
        atoms = results.pop("atoms")
        energy = results.pop("energy")
        forces = results.pop("forces")
        stress = results.pop("stress")

        atoms.set_velocities(at.get_velocities())
        calc = SPCalc(atoms,
                      energy=energy,
                      forces=forces,
                      stress=stress)
        calc.version = results.pop("version")
        atoms.calc = calc
        return atoms

# ========================================================================== #
    def _organize_pseudos(self, pseudos):
        """
        To have the pseudo well organized, we need to sort the pseudos
        """
        typat = []
        pseudolist = []
        for ityp in pseudos.keys():
            typat.append(ityp)
            pseudolist.append(pseudos[ityp])
        typat = np.array(typat)
        pseudolist = np.array(pseudolist)

        self.typat = typat
        znucl = symbols2numbers(typat)
        idx = np.argsort(znucl)

        # Reorder in increasing Z
        self.typat = typat[idx]
        self.pseudos = pseudolist[idx]

# ========================================================================== #
    def _remove_previous_run(self, stateprefix):
        """
        Little function to remove any trace of previous calculation
        """
        if os.path.exists(stateprefix + "abinit.abi"):
            os.remove(stateprefix + "abinit.abi")
        if os.path.exists(stateprefix + "abinit.abo"):
            os.remove(stateprefix + "abinit.abo")
        if os.path.exists(stateprefix + self.log):
            os.remove(stateprefix + self.log)
        if os.path.exists(stateprefix + "abinito_GSR.nc"):
            os.remove(stateprefix + "abinito_GSR.nc")
        if os.path.exists(stateprefix + "abinito_OUT.nc"):
            os.remove(stateprefix + "abinito_OUT.nc")
        if os.path.exists(stateprefix + "abinito_DEN"):
            os.remove(stateprefix + "abinito_DEN")
        if os.path.exists(stateprefix + "abinito_WF"):
            os.remove(stateprefix + "abinito_WF")
        if os.path.exists(stateprefix + "abinito_DDB"):
            os.remove(stateprefix + "abinito_DDB")
        if os.path.exists(stateprefix + "abinito_EIG"):
            os.remove(stateprefix + "abinito_EIG")
        if os.path.exists(stateprefix + "abinito_EBANDS.agr"):
            os.remove(stateprefix + "abinito_EBANDS.agr")

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        cmd = self.abinit_cmd
        cmd += ' --version'
        version = check_output(cmd, shell=True).decode('utf-8')
        msg = "True potential parameters:\n"
        msg += f"Abinit : {version}\n"
        dct = self.parameters
        msg += "parameters :\n"
        for key in dct.keys():
            msg += "   " + key + "  {0}\n".format(dct[key])
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
import shutil
from pathlib import Path

from ase import Atoms
import pytest

from ... import context  # noqa

from mlacs.mlip import SnapDescriptor, MliapDescriptor, LinearPotential
from mlacs.mlip import MomentTensorPotential, AceDescriptor, TensorpotPotential
from mlacs.mlip import UniformWeight
from mlacs.state import LammpsState


def generate_atoms():
    # Create_H2O
    positions = [[0.0, 0.0, 0.0],          # O
                 [0.9584, 0.0, 0.0],       # H
                 [-0.2396, 0.9273, 0.0]]   # H
    cell = [10.0, 10.0, 10.0]
    h2o = Atoms(symbols="OH2", positions=positions, cell=cell, pbc=True)

    # Create_H2O2
    positions = [[0.0, 0.0, 0.0],          # O
                 [1.45, 0.0, 0.0],         # O
                 [-0.78, 0.76, 0.0],       # H
                 [2.23, 0.76, 0.0]]        # H
    cell = [10.0, 10.0, 10.0]
    h2o2 = Atoms(symbols="H2O2", positions=positions, cell=cell, pbc=True)

    # Create_CH4
    positions = [[0.0, 0.0, 0.0],          # C
                 [0.63, 0.63, 0.63],       # H
                 [-0.63, -0.63, 0.63],     # H
                 [0.63, -0.63, -0.63],     # H
                 [-0.63, 0.63, -0.63]]      # H
    cell = [10.0, 10.0, 10.0]
    ch4 = Atoms(symbols="CH4", positions=positions, cell=cell, pbc=True)
    return [h2o, h2o2, ch4]


def generate_mlip(atoms, mlip_name):
    rcut = 5
    uni = UniformWeight(nthrow=10,
                        energy_coefficient=1.0,
                        forces_coefficient=1.0,
                        stress_coefficient=0.0)

    # SNAP
    if mlip_name == "SNAP_HOC.model":
        snap_params = {"twojmax": 4}
        snap_desc = SnapDescriptor(atoms=atoms,
                                   rcut=rcut,
                                   parameters=snap_params,
                                   model="linear")
        snap = LinearPotential(descriptor=snap_desc,
                               weight=uni, folder="SNAP")
        return snap

    # MLIAP
    if mlip_name == "MLIAP_HOC.model":
        snap_params = {"twojmax": 4}
        mliap_desc = MliapDescriptor(atoms=atoms,
                                     rcut=rcut,
                                     parameters=snap_params,
                                     model="linear",
                                     style="snap")
        mliap = LinearPotential(descriptor=mliap_desc,
                                weight=uni, folder="MLIAP")
        return mliap

    # MTP
    if mlip_name == "MTP_HOC.mtp":
        mtp_params = {"level": 6}
        mtp = MomentTensorPotential(atoms=atoms, mtp_parameters=mtp_params)
        return mtp
    # ACE
    if mlip_name == "ACE_HOC.yace":
        ace_desc = AceDescriptor(atoms, free_at_e={'C': 0, 'H': 0, 'O': 0},
                                 tol_e=100, tol_f=250, rcut=rcut)
        ace = TensorpotPotential(ace_desc, weight=uni, folder="ACE")
        return ace


@pytest.mark.parametrize("mlip_fn", ["SNAP_HOC.model", "MLIAP_HOC.model",
                                     "MTP_HOC.mtp", "ACE_HOC.yace"])
def test_lammpsstate_varntypat(root, mlip_fn):
    if not (Path(root) / "reference_files").exists():
        root = root.parents[0]  # Go up 1 folders to find reference_files
    if not (Path(root) / "reference_files").exists():
        root = root.parents[0]  # Go up another folder to find reference_files

    atoms = generate_atoms()
    mlip = generate_mlip(atoms, mlip_fn)
    fn = Path(root) / "reference_files" / "Pretrained-Potential" / mlip_fn
    state = LammpsState(temperature=500, dt=0.01, nsteps=5)
    folder = "MLIP"

    # The loading of MLIP to run_dynamics should be cleaned up
    if fn.suffix == ".model":  # SNAP or MLIAP
        coef = mlip.descriptor.get_coef(fn)
        if "SNAP" in str(fn):
            folder = "SNAP"
        else:
            folder = "MLIAP"
    elif fn.suffix == ".mtp":  # MTP
        folder = Path("MTP")
        folder.mkdir(exist_ok=True)
        mtp_fn = folder / fn.name
        shutil.copy(fn, mtp_fn)
        coef = fn
        with open(folder / "mlip.ini", "w") as f:
            f.write(f"mtp-filename    {mtp_fn.absolute()}\n")
            f.write("select          FALSE")
    elif fn.suffix == ".yace":  # ACE
        coef = fn
        folder = Path("ACE")
        folder.mkdir(exist_ok=True)
        shutil.copy(fn, folder / "ACE.yace")
    mlip.next_coefs(coef)
    for molecule in atoms:
        state.run_dynamics(supercell=molecule,
                           pair_style=mlip.pair_style,
                           pair_coeff=mlip.pair_coeff,
                           elements=mlip.get_elements())
    shutil.rmtree(folder)
    shutil.rmtree("Trajectory")


@pytest.mark.parametrize("mlip_fn", ["SNAP_HOC.model", "MLIAP_HOC.model",
                                     "ACE_HOC.yace"])
def test_predict_varntypat(root, mlip_fn):
    """
    Test that a MLIAP/SNAP/MTP/ACE can use the predict function
    for a variable natom/ntypat
    """
    if not (Path(root) / "reference_files").exists():
        root = root.parents[0]  # Go up 1 folders to find reference_files
    if not (Path(root) / "reference_files").exists():
        root = root.parents[0]  # Go up another folder to find reference_files

    atoms = generate_atoms()
    mlip = generate_mlip(atoms, mlip_fn)
    fn = Path(root) / "reference_files" / "Pretrained-Potential" / mlip_fn
    folder = "MLIP"

    # The loading of MLIP to run_dynamics should be cleaned up
    if fn.suffix == ".model":  # SNAP or MLIAP
        coef = mlip.descriptor.get_coef(fn)
        if "SNAP" in str(fn):
            folder = "SNAP"
        else:
            folder = "MLIAP"
    elif fn.suffix == ".mtp":  # MTP
        folder = Path("MTP")
        folder.mkdir(exist_ok=True)
        mtp_fn = folder / fn.name
        shutil.copy(fn, mtp_fn)
        coef = fn
        with open(folder / "mlip.ini", "w") as f:
            f.write(f"mtp-filename    {mtp_fn.absolute()}\n")
            f.write("select          FALSE")
    elif fn.suffix == ".yace":  # ACE
        coef = fn
        folder = Path("ACE")
        folder.mkdir(exist_ok=True)
        shutil.copy(fn, folder / "ACE.yace")
    mlip.next_coefs(coef)

    mlip.predict(atoms)
    shutil.rmtree(folder)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk

from ... import context  # noqa
from mlacs.ti import UFLiquidState, ThermodynamicIntegration


@pytest.fixture
def expected_folder(expected_folder_base):
    return ["ThermoInt"]


@pytest.fixture
def expected_files(expected_files_base):
    return []


def test_einstein_onestate(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    pair_style = "eam/alloy"
    potfold = (root / "reference_files").as_posix()
    pair_coeff = [f"* * {potfold}/Cu01.eam.alloy Cu"]
    temperature = 300

    nsteps = 5

    state = UFLiquidState(atoms,
                          pair_style,
                          pair_coeff,
                          temperature,
                          nsteps=nsteps,
                          nsteps_eq=nsteps,
                          nsteps_msd=nsteps,
                          nsteps_averagin=nsteps)
    ti = ThermodynamicIntegration(state)
    ti.run()


def test_einstein_twostate(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    pair_style = "eam/alloy"
    potfold = (root / "reference_files").as_posix()
    pair_coeff = [f"* * {potfold}/Cu01.eam.alloy Cu"]

    nsteps = 5

    states = []
    for t in [300, 500]:
        states.append(UFLiquidState(atoms,
                      pair_style,
                      pair_coeff,
                      t,
                      0.0,
                      nsteps=nsteps,
                      nsteps_eq=nsteps,
                      nsteps_msd=nsteps,
                      nsteps_averagin=nsteps))
    ti = ThermodynamicIntegration(states)
    ti.run()


def test_einstein_twoinstance(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    pair_style = "eam/alloy"
    potfold = (root / "reference_files").as_posix()
    pair_coeff = [f"* * {potfold}/Cu01.eam.alloy Cu"]
    temperature = 300

    nsteps = 5

    state = UFLiquidState(atoms,
                          pair_style,
                          pair_coeff,
                          temperature,
                          nsteps=nsteps,
                          nsteps_eq=nsteps,
                          nsteps_msd=nsteps,
                          nsteps_averagin=nsteps)
    ti = ThermodynamicIntegration(state, 2)
    ti.run()



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import numpy as np
from abc import ABC, abstractmethod

from ase.atoms import Atoms
from ase.units import GPa

from ..core import Manager
from ..utilities import compute_correlation, create_link
from .weights import UniformWeight


# ========================================================================== #
# ========================================================================== #
class MlipManager(Manager, ABC):
    """
    Parent Class for the management of Machine-Learning Interatomic Potential
    """
    def __init__(self,
                 descriptor,
                 weight=None,
                 folder='MLIP',
                 **kwargs):

        Manager.__init__(self, folder=folder, **kwargs)

        self.descriptor = descriptor
        self.descriptor.workdir = self.workdir
        self.descriptor.folder = self.folder

        self.amat_e = None
        self.amat_f = None
        self.amat_s = None

        self.ymat_e = None
        self.ymat_f = None
        self.ymat_s = None

        self.natoms = []

        self.fit_res = None

        self.weight = weight
        if self.weight is None:
            self.weight = UniformWeight()

        self.weight.workdir = self.workdir
        self.weight.folder = self.folder
        self.weight.subfolder = self.subfolder

        self.nconfs = 0

        # Some initialization for sampling interface
        self.model_post = None
        self.atom_style = "atomic"

        self.can_use_weight = False

# ========================================================================== #
    def update_matrices(self, atoms):
        """
        Update database, feature matrix and label vector with new configs.

        Parameters
        -----

        atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms`
            New configuration(s) to be added to the database.

        Notes
        -----

        Feature matrix `amat_i` and label vector `ymat_i` are separated into
        blocks i=e,f,s, representing energy, forces and stresses, respectively.

        Attribute dimensions after update
            - `self.amat_e`: ndarray of shape (N, K)
                Block matrix representing the descriptors
            - `self.amat_f`: ndarray of shape (3*Nat*N, K)
                Block matrix representing the gradient of descr. wrt positions
            - `self.amat_s`: ndarray of shape (6*N, K)
                Block matrix representing the gradient of descr. wrt strains
            - `self.ymat_e`: ndarray of shape (N,)
                Label vector representing the energies
            - `self.ymat_f`: ndarray of shape (3*Nat*N,)
                Label vector representing the forces
            - `self.ymat_s`: ndarray of shape (6*N,)
                Label vector representing the stresses

        Where:
            - `K` is the number of descriptor components (features)
            - `N` is the number of configurations in database (after update)
            - `Nat` is the number of atoms in each cell
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.weight.update_database(atoms)

        self.descriptor.workdir = self.workdir
        self.descriptor.folder = self.folder
        amat_all = self.descriptor.compute_descriptors(atoms)

        energy = np.array([at.get_potential_energy() for at in atoms])
        forces = np.array([at.get_forces() for at in atoms]).flatten()
        stress = np.array([at.get_stress() for at in atoms]).flatten()
        nat = np.array([len(at) for at in atoms])

        for amat in amat_all:
            if self.amat_e is None:
                self.amat_e = amat["desc_e"]
                self.amat_f = amat["desc_f"]
                self.amat_s = amat["desc_s"]
            else:
                self.amat_e = np.r_[self.amat_e, amat["desc_e"]]
                self.amat_f = np.r_[self.amat_f, amat["desc_f"]]
                self.amat_s = np.r_[self.amat_s, amat["desc_s"]]

        if self.ymat_e is None:
            self.ymat_e = energy
            self.ymat_f = forces
            self.ymat_s = stress
            self.natoms = nat
        else:
            self.ymat_e = np.r_[self.ymat_e, energy]
            self.ymat_f = np.r_[self.ymat_f, forces]
            self.ymat_s = np.r_[self.ymat_s, stress]
            self.natoms = np.append(self.natoms, [nat])

        self.nconfs += len(atoms)

# ========================================================================== #
    @abstractmethod
    def train_mlip(self):
        """
        """
        raise NotImplementedError

# ========================================================================== #
    @abstractmethod
    def predict(self, atoms, coef=None):
        """
        Function that gives the e, f, s
        """
        raise NotImplementedError

# ========================================================================== #
    def read_parent_mlip(self, traj):
        """
        Get a list of all the mlip that have generated a conf in traj
        and get the coefficients of all these mlip
        """
        parent_mlip = []
        mlip_coef = []
        prefix = self.descriptor.prefix
        directory = self.descriptor.subdir

        # Check that this simulation and the previous one use the same mlip
        fn_descriptor = self.subdir / f"{prefix}.descriptor"
        with open(fn_descriptor, "r") as f:
            previous_mlip = f.read()

        if not previous_mlip == self.descriptor.get_mlip_params():
            err = "The MLIP.descriptor from {fn_descriptor} seems different "
            err += "to the one you have in this simulation. If you want a "
            err += "new mlip: Rerun MLACS with DatabaseCalculator and "
            err += "OtfMlacs.keep_tmp_files=True on your traj"
            raise ValueError(err)

        # Make the MBAR variable Nk and mlip_coef
        for state in traj:
            for conf in state:
                if "parent_mlip" not in conf.info:  # Initial or training
                    continue
                else:  # A traj
                    # GA: Not sure if this is absolute or relative
                    model = conf.info['parent_mlip']
                    directory = Path(model)
                    if not directory.exists:
                        # GA: If the files have been moved,
                        #    it wont be possible to restart the calculation.
                        #    However, one might want to restart a calculation
                        #    on a different machine than the one it started on.
                        #    TODO: Get directories by inspection instead.
                        # ON: I agree
                        err = "Some parent MLIP are missing. "
                        err += "Rerun MLACS with DatabaseCalculator and "
                        err += "OtfMlacs.keep_tmp_files=True on your traj"
                        raise FileNotFoundError(err)
                    if model not in parent_mlip:  # New state
                        parent_mlip.append(model)
                        fn = directory / f"{prefix}.model"
                        coef = self.descriptor.get_coef(filename=fn)
                        mlip_coef.append(coef)
        return parent_mlip, np.array(mlip_coef)

# ========================================================================== #
    def next_coefs(self, mlip_coef):
        """
        Update MLACS just like train_mlip, but without actually computing
        the coefficients
        """
        self.weight.subfolder = self.subfolder
        self.descriptor.subfolder = self.subfolder

        self.coefficients = mlip_coef

        # GA: Passing names like this is a bit shady. TODO: clean up.
        mlip_fn = self.descriptor.write_mlip(mlip_coef)
        _, __ = self.weight.compute_weight(mlip_coef,
                                           self.predict,
                                           docalc=False)
        prefix = self.descriptor.prefix
        desc_fn = self.subdir/f"{prefix}.descriptor"
        if not Path(desc_fn).exists():
            self.descriptor._write_mlip_params()

        # GA: Not sure why we need to create a link here.
        create_link(self.subsubdir/mlip_fn, self.subdir/f"{prefix}.model")

# ========================================================================== #
    @Manager.exec_from_workdir
    def test_mlip(self, testset):
        """
        """
        calc = self.get_calculator()

        ml_e = []
        ml_f = []
        ml_s = []
        dft_e = []
        dft_f = []
        dft_s = []
        for at in testset:
            mlat = at.copy()
            mlat.calc = calc
            e = mlat.get_potential_energy() / len(mlat)
            f = mlat.get_forces().flatten()
            s = mlat.get_stress()

            ml_e.append(e)
            ml_f.extend(f)
            ml_s.extend(s)

            e = at.get_potential_energy() / len(at)
            f = at.get_forces().flatten()
            s = at.get_stress()

            dft_e.append(e)
            dft_f.extend(f)
            dft_s.extend(s)

        dft_e = np.array(dft_e)
        dft_f = np.array(dft_f)
        dft_s = np.array(dft_s)
        ml_e = np.array(ml_e)
        ml_f = np.array(ml_f)
        ml_s = np.array(ml_s)

        rmse_e, mae_e, rsq_e = compute_correlation(np.c_[dft_e, ml_e])
        rmse_f, mae_f, rsq_f = compute_correlation(np.c_[dft_f, ml_f])
        rmse_s, mae_s, rsq_s = compute_correlation(np.c_[dft_s, ml_s] / GPa)

        nat = np.array([len(at) for at in testset]).sum()
        msg = "number of configurations for training: " + \
              f"{len(testset)}\n"
        msg += "number of atomic environments for training: " + \
               f"{nat}\n"

        # Prepare message to the log
        msg += f"RMSE Energy    {rmse_e:.4f} eV/at\n"
        msg += f"MAE Energy     {mae_e:.4f} eV/at\n"
        msg += f"RMSE Forces    {rmse_f:.4f} eV/angs\n"
        msg += f"MAE Forces     {mae_f:.4f} eV/angs\n"
        msg += f"RMSE Stress    {rmse_s:.4f} GPa\n"
        msg += f"MAE Stress     {mae_s:.4f} GPa\n"
        msg += "\n"

        header = f"rmse: {rmse_e:.5f} eV/at,    " + \
                 f"mae: {mae_e:.5f} eV/at\n" + \
                 " True Energy           Predicted Energy"
        np.savetxt("TestSet-Energy_comparison.dat",
                   np.c_[dft_e, ml_e],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"rmse: {rmse_f:.5f} eV/angs   " + \
                 f"mae: {mae_f:.5f} eV/angs\n" + \
                 " True Forces           Predicted Forces"
        np.savetxt("TestSet-Forces_comparison.dat",
                   np.c_[dft_f, ml_f],
                   header=header, fmt="%25.20f  %25.20f")
        header = f"rmse: {rmse_s:.5f} GPa       " + \
                 f"mae: {mae_s:.5f} GPa\n" + \
                 " True Stress           Predicted Stress"
        np.savetxt("TestSet-Stress_comparison.dat",
                   np.c_[dft_s, ml_s] / GPa,
                   header=header, fmt="%25.20f  %25.20f")
        return msg

# ========================================================================== #
    @property
    def pair_style(self):
        return self._get_pair_style()

    @property
    def pair_coeff(self):
        return self._get_pair_coeff()

# ========================================================================== #
    def _get_pair_style(self):
        self.descriptor.folder = self.folder
        return self.descriptor.get_pair_style()

# ========================================================================== #
    def _get_pair_coeff(self):
        self.descriptor.folder = self.folder
        return self.descriptor.get_pair_coeff()

# ========================================================================== #
    def get_elements(self):
        return self.descriptor.elements


# ========================================================================== #
# ========================================================================== #
class SelfMlipManager(MlipManager):
    """
    Mlip manager for model that both compute the descriptor and takes
    care of the regression (ex. MTP, POD)
    """
    def __init__(self,
                 descriptor,
                 weight=None,
                 **kwargs):
        MlipManager.__init__(self, descriptor, weight=weight, **kwargs)
        self.configurations = []
        self.natoms = []

# ========================================================================== #
    def update_matrices(self, atoms):
        """
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        nat = np.array([len(at) for at in atoms], dtype=int)

        self.configurations.extend(atoms)
        self.natoms = np.append(self.natoms, nat)
        self.natoms = np.array(self.natoms, dtype=int)
        self.nconfs += len(atoms)



"""
// Copyright (C) 2022-2024 MLACS group (RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np

from scipy.spatial import distance

from ase.atoms import Atoms
from ase.calculators.singlepoint import SinglePointCalculator as SPC

from ..utilities import interpolate_points as intpts


class PathAtoms:
    """
    Base class for managing transiton state.

    Parameters
    ----------
    images: :class:`list` or `PathAtoms`
        mlacs.PathAtoms or list of ase.Atoms object.
        The list contain initial and final configurations of the reaction path.

    xi: :class:`numpy.array` or `float`
        Value of the reaction coordinate for the constrained MD.
        Default ``None``

    mode: :class:`float` or :class:`string`
        Value of the reaction coordinate or sampling mode
            - ``saddle`` return the saddle point.
            - ``float`` sampling at a precise coordinate.
            - ``rdm`` randomly return the coordinate of a splined images.
            - ``rdm_true`` randomly return the coordinate of an images.
            - ``rdm_memory`` homogeneous sampling of splined images.
            - ``gaussian`` Bayesian Inference sampling.
        Default ``saddle``

    interval: :class:`list`
        Define the interval limits to sample the reaction coordinate.
        Take only two values [min, max].
        Default ``[0, 1]``

    fixcom: :class: `bool`
        Include the center of mass displacement along the reaction coordinate.
        Default ``True``

    """
    def __init__(self,
                 images,
                 xi=None,
                 mode="saddle",
                 interval=None,
                 fixcom=True,
                 **kwargs):

        self.mode = mode
        self.images = images
        if len(images) < 2:
            msg = 'You need at least a starting and a final configuration '
            msg += 'to define a Transition Path !'
            raise TypeError(msg)
        self.fixcom = fixcom

        self._xi = xi
        self._tmp_xi = xi
        self._memxi = interval
        self._splR = None
        self._splined = None
        self._splprec = 1001
        if self._memxi is None:
            self._memxi = [0.0, 1.0]
        self._gpi = None

    @property
    def images(self):
        """Get true images"""
        return self._images

    @property
    def initial(self):
        """Get first image"""
        return self._images[0]

    @property
    def final(self):
        """Get last image"""
        return self._images[-1]

    @images.setter
    def images(self, img):
        self._images = img

    @property
    def imgxi(self):
        return np.linspace(0, 1, len(self.images))

    @property
    def imgR(self):
        return np.array([a.positions for a in self.images])

    @property
    def imgE(self):
        _imgE = [a.get_potential_energy() for a in self.images]
        return np.array(_imgE, dtype=float)

    @property
    def imgC(self):
        return np.array([a.cell.array for a in self.images])

    @property
    def nreplica(self):
        return len(self.images)

    @property
    def masses(self):
        """Get the effective masses of particles"""
        pos = np.transpose(self.imgR, (1, 0, 2))
        meff = np.array([np.max(distance.cdist(d, d, "euclidean"))
                         for d in pos])
        return meff / np.max(meff)

    @property
    def update(self):
        """
        Update the reaction coordinate.
        """
        if self._xi is not None:
            self._tmp_xi = self._xi
            return self._xi

        def find_dist(_l):
            m = []
            _l.sort()
            for i, val in enumerate(_l[1:]):
                m.append(np.abs(_l[i+1] - _l[i]))
            i = np.array(m).argmax()
            return _l[i+1], _l[i]

        mode = self.mode
        if isinstance(mode, (float, list, np.ndarray)):
            self._tmp_xi = mode
            return mode
        elif mode == 'saddle':
            # RB : We can't evaluate the energy at the initialisation.
            if self._tmp_xi is None:
                self._tmp_xi = 0.5
                return self._tmp_xi
            self._tmp_xi, _ = self.saddle_point
            return self._tmp_xi
        elif mode == 'rdm':
            i, f = tuple(self._memxi)
            self._tmp_xi = np.random.uniform(i, f)
            return self._tmp_xi
        elif mode == 'rdm_memory':
            i, f = find_dist(self._memxi)
            self._tmp_xi = np.random.uniform(i, f)
            self._memxi.append(self._tmp_xi)
            return self._tmp_xi
        elif mode == 'rdm_true':
            def cond(x, i, f): return x >= i and x <= f
            i, f = tuple(self._memxi)
            xiint = list(filter(lambda x: cond(x, i, f), self.imgxi))
            self._tmp_xi = np.random.choice(xiint)
            return self._tmp_xi
        elif mode == 'gaussian':
            self._tmp_xi = self._gaussian_process()
            return self._tmp_xi
        else:
            msg = 'The reaction coordinate is not defined.'
            msg += 'You need to define `xi` or the `mode` to find it !'
            raise TypeError(msg)

    @property
    def xi(self):
        """
        Get the reaction coordinate.
        """
        if self._xi is not None:
            return self._xi
        elif self._tmp_xi is None:
            return self.update
        else:
            return self._tmp_xi

    @xi.setter
    def xi(self, xi):
        self._xi = xi

    @property
    def splined(self):
        """Get splined images"""
        self._splined = self.get_splined_atoms(self.xi)
        return self._splined

    @property
    def splR(self):
        self.set_splined_matrices(self.xi)
        return self._splR[:, 0:3]

    @property
    def splDR(self):
        self.set_splined_matrices(self.xi)
        if self.fixcom:
            return self._splR[:, 9:12]
        return self._splR[:, 3:6]

    @property
    def splD2R(self):
        self.set_splined_matrices(self.xi)
        if self.fixcom:
            return self._splR[:, 12:15]
        return self._splR[:, 6:9]

    @property
    def splE(self):
        # RB : The derivatives at xi=0 or 1 is not always null.
        # return np.r_[intpts(self.imgxi, self.imgE, self.xi, 0, border=1)]
        _splE = self.set_splined_energy(self.xi)
        return _splE

    @property
    def saddle_point(self):
        x = np.linspace(0, 1, self._splprec)
        # RB : The derivatives at xi=0 or 1 is not always null.
        # y = np.r_[intpts(self.imgxi, self.imgE, x, 0, border=1)]
        y = self.set_splined_energy(x)
        return x[y.argmax()], max(y)

    @property
    def saddle(self):
        """Get the saddle point image"""
        _xi = self._xi
        xi, _ = self.saddle_point
        self._saddle = self.get_splined_atoms(xi)
        self._xi = _xi
        return self._saddle

    @property
    def splC(self):
        _imgC = self.imgC.reshape((self.nreplica, 9)).T
        if isinstance(self.xi, (float, int)):
            shape = (3, 3)
            _splC = np.zeros(9)
        else:
            shape = (len(self.xi), 3, 3)
            _splC = np.zeros((9, len(self.xi)))
        for i in range(9):
            _splC[i] = intpts(self.imgxi, _imgC[i], self.xi, 0, border=1)
        return _splC.T.reshape(shape)

    def get_splined_atoms(self, xi=None):
        """
        Return splined Atoms objects at the xi coordinates.
        """
        if xi is None:
            xi = self.update

        def set_atoms(X, C, M, R, DR, D2R):
            Z = self.images[0].get_atomic_numbers()
            at = Atoms(numbers=Z, positions=R, cell=C)
            at.set_pbc(True)
            at.set_array('first_derivatives', DR)
            at.set_array('second_derivatives', D2R)
            at.set_array('effective_masses', M)
            at.info['reaction_coordinate'] = X
            return at

        if isinstance(xi, float):
            splat = set_atoms(xi, self.splC, self.masses,
                              self.splR, self.splDR, self.splD2R)
            calc = SPC(atoms=splat, energy=self.splE)
            splat.calc = calc
            return splat

        splat = []
        for i, (x, c, e) in enumerate(zip(xi, self.splC, self.splE)):
            at = set_atoms(x, c, self.masses, self.splR[:, :, i],
                           self.splDR[:, :, i], self.splD2R[:, :, i])
            calc = SPC(atoms=at, energy=e)
            at.calc = calc
            splat.append(at)
        return splat

    def set_splined_energy(self, xi=None):
        if xi is None:
            xi = np.linspace(0, 1, self._splprec)
        # RB : The derivatives at xi=0 or 1 is not always null.
        # y = np.r_[intpts(self.imgxi, self.imgE, x, 0, border=1)]
        y = np.r_[intpts(self.imgxi, self.imgE, xi, 0)]
        return np.array(y, dtype=float)

    def set_splined_matrices(self, xi=None):
        """
        Compute a 1D CubicSpline interpolation from a NEB calculation.
        The function also set up the lammps data file for a constrained MD.
            - Three first columns: atomic positons at reaction coordinate xi.
            - Three next columns:  normalized atomic first derivatives at
                reaction coordinate xi, with the corrections of the COM.
            - Three last columns:  normalized atomic second derivatives at
                reaction coordinate xi.
        """
        if xi is None:
            xi = self.update

        N = len(self.images[0])
        if isinstance(xi, float):
            _nim = 1
            self._splR = np.zeros((N, 15))
        else:
            _nim = len(xi)
            self._splR = np.zeros((N, 15, _nim))

        # Spline interpolation of the referent path and calculation of
        # the path tangent and path tangent derivate
        for i in range(N):
            self._splR[i, 0:3] = np.r_[[intpts(self.imgxi, self.imgR[:, i, j],
                                        xi, 0) for j in range(3)]]
            self._splR[i, 3:6] = np.r_[[intpts(self.imgxi, self.imgR[:, i, j],
                                        xi, 1) for j in range(3)]]
            self._splR[i, 6:9] = np.r_[[intpts(self.imgxi, self.imgR[:, i, j],
                                        xi, 2) for j in range(3)]]

        if self.fixcom:
            self._com_corrections()
        return self._splR.round(8)

    def _com_corrections(self):
        """
        Correction of the path tangent to have zero center of mass
        """
        N = len(self.images[0])
        der = self._splR[:, 3:6]
        der2 = self._splR[:, 6:9]
        com = np.array([np.average(der[:, i]) for i in range(3)])
        n = 0
        for i in range(N):
            n += np.sum([(der[i, j] - com[j])**2 for j in range(3)])
        n = np.sqrt(n)
        for i in range(N):
            self._splR[i, 9:12] = np.r_[[(der[i, j] - com[j]) / n
                                        for j in range(3)]]
            self._splR[i, 12:15] = np.r_[[der2[i, j] / n / n
                                         for j in range(3)]]

    def _gaussian_process(self):
        """
        Run Gaussian process regressor to do Bayesian inference.
        """
        from mlacs.core import GaussianProcessInterface as GPI
        from sklearn.gaussian_process.kernels import (RBF,
                                                      ConstantKernel as C)

        # RB : We can't evaluate the energy at the initialisation.
        if self._tmp_xi is None:
            x = np.r_[[self.imgxi[0], self.imgxi[-1]]]
            y = np.r_[[self.imgE[0], self.imgE[-1]]]

            self._gpi = GPI(kernel=RBF()*C(),
                            gp_parameters=dict(normalize_y=False))
            self._gpi.add_new_data(x, y)

            self._gp_minE = self.set_splined_energy()
            self._gp_maxE = self.set_splined_energy()
            self._gp_error = self._gp_maxE - self._gp_minE
            self._gp_error_max = max(np.abs(self._gp_error))

            i, f = tuple(self._memxi)
            return np.random.uniform(i, f)

        error = self.set_splined_energy() - self._gp_minE
        if max(np.abs(error)) > self._gp_error_max:
            self._gp_maxE = self.set_splined_energy()
        error = self._gp_maxE - self.set_splined_energy()
        if max(np.abs(error)) > self._gp_error_max:
            self._gp_minE = self.set_splined_energy()
        self._gp_error = self._gp_maxE - self._gp_minE
        self._gp_error_max = max(np.abs(self._gp_error))

        if len(self._gpi.y) <= 3:
            self._gpi.add_new_data(np.r_[[self.xi]], self.splE)

            i, f = tuple(self._memxi)
            return np.random.uniform(i, f)
        else:
            self._gpi.add_new_data(np.r_[[self.xi]], self.splE)

        x = np.linspace(0, 1, self._splprec)
        xi = self._gpi.x.reshape(1, -1)[0]
        alpha = np.array(intpts(x, self._gp_error, xi, 0), dtype=float)

        self._gpi.alpha_ = alpha
        self._gpi.train()
        mean, std = self._gpi.predict(x, True)

        std = np.max(std, axis=0) - np.min(std, axis=0)
        return x[np.argmax(std)]



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import logging
import numpy as np
from ase.units import GPa
try:
    # With the annoying mandatory warning from mbar, we have to initialize
    # the log here otherwise the log doesn't work
    # I have to see how to handle this in a better way.
    # This might be an indication of needing to redo the logging
    logging.basicConfig(level=logging.INFO, format='%(message)s')
    logger = logging.getLogger("pymbar")
    logger.setLevel(logging.ERROR)
    from pymbar import MBAR
    logger.setLevel(logging.INFO)
    ispymbar = True
except ModuleNotFoundError:
    ispymbar = False

from ase.atoms import Atoms
from ase.units import kB
from ..core.manager import Manager
from .weighting_policy import WeightingPolicy


default_parameters = {"solver": "L-BFGS-B",
                      "scale": 1.0,
                      "start": 2,
                      }


# ========================================================================== #
# ========================================================================== #
class MbarManager(WeightingPolicy):
    """
    Computation of weight according to the multistate Bennett acceptance
    ratio (MBAR) method for the analysis of equilibrium samples from multiple
    arbitrary thermodynamic states.

    Parameters
    ----------
    solver: :class:`str`
        Define type of solver for pymbar
        Default L-BFGS-B

    scale: :class:`float`
        Imposes weights for the new configurations.
        Default 1.0

    start: :class:`int`
        Step to start weight computation.
        At least 2 since you need two potentials to compare them.
        Default 2

    database: :class:`ase.Trajectory`
        Initial database (optional)
        Default :class:`None`

    weight: :class:`list` or :class:`str`
        If you use an initial database, it needs weight.
        Can a list or an np.array of values or a file.
        Default :class:`None`
    """

    def __init__(self, parameters=dict(),  energy_coefficient=1.0,
                 forces_coefficient=1.0, stress_coefficient=1.0,
                 **kwargs):

        if not ispymbar:
            msg = "You need pymbar installed to use the MBAR manager"
            raise ModuleNotFoundError(msg)

        WeightingPolicy.__init__(self,
                                 energy_coefficient=energy_coefficient,
                                 forces_coefficient=forces_coefficient,
                                 stress_coefficient=stress_coefficient,
                                 **kwargs)

        if self.database is None:
            self.database = []
        self.parameters = default_parameters
        self.parameters.update(parameters)
        self.Nk = []
        self.train_mlip = False
        self.mlip_coef = []
        self.ukn = []

        self._newddb = []
        self._nstart = self.parameters['start']
        if self._nstart <= 1:
            msg = 'The "start" variable has to be higher than 2.\n'
            msg += 'You need at least two potentials to compare them.'
            raise ValueError(msg)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef, predict, docalc=True):
        """
        Save the MLIP coefficients and compute the Weight
        Compute the matrice Ukn of partition fonctions of shape [ndesc, nconf]
        according to the given predict(desc, coef)
        """
        # Update the variables
        if coef is not None:
            self.mlip_coef.append(coef)

        self.train_mlip = True
        self.database.extend(self._newddb)
        self.nconfs = len(self.database)
        if 0 == len(self.mlip_coef):
            self.Nk = np.r_[self.nconfs]
        else:
            self.Nk = np.append(self.Nk, [len(self._newddb)])
        self._newddb = []

        if not docalc:
            return "_", "_"

        # Calculate ukn
        ukn = np.zeros([len(self.mlip_coef), len(self.database)])
        for idx, mlip_coef in enumerate(self.mlip_coef):
            # Calculate the MLIP Energy and Pressure according to the new coef
            mlip_E, mlip_F, mlip_S = predict(self.database, mlip_coef)
            mlip_P = []
            for s in mlip_S:
                mlip_P.append(-np.sum(s[:3])/3)

            ukn[idx] = self._get_ukn(energy=mlip_E,
                                     pressure=mlip_P)
        self.ukn = ukn
        squared_ukn = np.zeros([len(self.ukn), len(self.ukn[-1])])
        for i in range(len(self.ukn)):
            for j in range(len(self.ukn[i])):
                squared_ukn[i, j] = self.ukn[i][j]

        # Finally, calculate weight
        fname = "MLIP.weight"
        header = ''
        if self._nstart <= len(self.mlip_coef):
            weight = self._compute_weight(squared_ukn)
            self.weight = weight
            neff = self.get_effective_conf()

            header += "Using MBAR weighting\n"
            header += f"Effective number of configurations: {neff:10.5f}\n"

            filepath = Path(fname)
            if filepath.exists():
                filepath.unlink()
            np.savetxt(fname, self.weight,
                       header=header, fmt="%25.20f")

            header += "Number of uncorrelated snapshots for each k state:\n"
            header += np.array2string(np.array(self.Nk, 'int')) + "\n"

        else:  # If there isn't enough coef, use UniformWeight
            self.weight = np.ones(len(self.database))/len(self.database)
        return header, "MLIP.weight"

# ========================================================================== #
    def init_weight(self):
        """
        Initialize the weight matrice with W = scale * 1/N.
        """
        return super().init_weight(scale=self.parameters['scale'])

# ========================================================================== #
    def update_database(self, atoms):
        """
        Update the database.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self._newddb.extend(atoms)
        if self.matsize is None:
            self.matsize = []
        self.matsize.extend([len(a) for a in atoms])

        # Sanity Check
        for at in atoms:
            if 'info_state' not in at.info:
                msg = "Atoms don't have 'info_state' for MBAR\n"
                msg += "To use mbar, look at the new traj file with 2 confs."
                msg += "Copy its info['info_state'] and add it to all the "
                msg += "atoms in traj."
                raise ValueError(msg)

# ========================================================================== #
    def _get_ukn(self, energy, pressure):
        """
        Compute Ukn matrices. u[k,n] -> k = mlip, n=conf
        """
        P, V, T = self._get_ensemble_info(pressure)
        PV = P*V / [len(at) for at in self.database]
        ukn = (energy + PV) / (kB * T)
        return ukn

# ========================================================================== #
    def _get_ensemble_info(self, pressure):
        """
        Read the ddb info state and returns arrays of P, dV, T.

        For now, only NVT and NPT are implemented.
        NVT : Aimed T, Instantaneous P, Constant V
        NPT : Aimed T, Constant P from the MLIP, Instantaneous V
        -----------------------------------------------
        NVE : Instantaneous T, No P, No V
        uVT/uPT : NVT/NPT + Constant u, Instantaneous N
        """
        P, V, T = [], [], []
        for idx, at in enumerate(self.database):
            info = at.info['info_state']
            ens = info['ensemble']
            if ens == "NVT":
                T = np.append(T, at.info['info_state']['temperature'])
                P = np.append(P, pressure[idx])
                V = np.append(V, at.get_volume())
            elif ens == "NPT":
                T = np.append(T, at.info['info_state']['temperature'])
                tmp_P = at.info['info_state']['pressure']
                P = np.append(P, tmp_P * GPa)  # From GPa to eV/ang**3
                V = np.append(V, at.get_volume())
            else:
                msg = "Only NVT and NPT are implemented in MLACS for now"
                raise NotImplementedError(msg)
        return P, V, T

# ========================================================================== #
    def _compute_weight(self, sq_ukn):
        """
        Uses pymbar.MAR() class.

        [1] Shirts MR and Chodera JD. Statistically optimal analysis of
        samples from multiple equilibrium states.
        J. Chem. Phys. 129:124105, 2008.  http://dx.doi.org/10.1063/1.2978177
        """
        mbar = MBAR(sq_ukn, self.Nk,
                    solver_protocol=[{'method': self.parameters['solver']}])
        weight = mbar.weights()[:, -1]
        return weight



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ase.atoms import Atoms
from ase.calculators.lammpsrun import LAMMPS
from ase.calculators.singlepoint import SinglePointCalculator

from .mlip_manager import MlipManager


# ========================================================================== #
# ========================================================================== #
class DeltaLearningPotential(MlipManager):
    """
    Parameters
    ----------
    model: :class:`MlipManager`
        The MLIP model to train on the difference between the true energy
        and the energy of a LAMMPS reference model.

    pair_style: :class:`str` or :class:`list` of :class:`str`
        The pair_style of the LAMMPS reference potential.
        If only one pair style is used, can be set as a :class:`str`.
        If an overlay of pair style is used, this input as to be a
        :class:`list` of :class:`str` of the pair_style.
        For example :

    pair_coeff: :class:`list` of :class:`str`
        The pair_coeff of the LAMMPS reference potential.

    folder: :class:`str` (optional)
        The root for the directory in which the MLIP are stored.
        Default 'MLIP'

    Examples
    --------

    >>> from ase.io import read
    >>> confs = read('Trajectory.traj', index=':')
    >>>
    >>> from mlacs.mlip import SnapDescriptor, LinearPotential
    >>> desc = SnapDescriptor(confs[0], rcut=6.2, parameters=dict(twojmax=6))
    >>> mlip = LinearPotential(desc)
    >>>
    >>> from mlacs.mlip import DeltaLearningPotential
    >>> ps = ['sw', 'zbl 3.0 4.0']
    >>> pc = [['* * Si.sw Si'], ['* * 14 14']]
    >>> dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=pc)
    >>> dmlip.update_matrices(confs)
    >>> dmlip.train_mlip()
    """
    def __init__(self,
                 model,
                 pair_style,
                 pair_coeff,
                 model_post=None,
                 atom_style="atomic",
                 folder=None,
                 **kwargs):

        if folder != model.folder:
            if folder is not None:
                model.folder = folder
            else:
                folder = model.folder

        self.model = model
        weight = self.model.weight

        if not isinstance(pair_style, list):
            pair_style = [pair_style]

        self.ref_pair_style = pair_style
        self.ref_pair_coeff = pair_coeff

        self.ref_model_post = model_post
        self.model_post = model_post
        self.ref_atom_style = atom_style
        self.atom_style = atom_style

        MlipManager.__init__(self, self.model.descriptor, weight,
                             folder=folder, **kwargs)

        self._ref_e = None
        self._ref_f = None
        self._ref_s = None

# ========================================================================== #

    def _set_directories(self):
        self.model.workdir = self.workdir
        self.model.folder = self.folder
        self.model.subfolder = self.subfolder
        self.model.descriptor.workdir = self.workdir
        self.model.descriptor.folder = self.folder
        self.model.descriptor.subfolder = self.subfolder
        self.model.weight.workdir = self.workdir
        self.model.weight.folder = self.folder
        self.model.weight.subfolder = self.subfolder

# ========================================================================== #
    def get_ref_pair_style(self, lmp=False):
        """
        Return self.ref_pair_style which is an array.
        If lmp=True, it returns it formatted as a lammps input.
        """
        self._set_directories()
        if not lmp:
            return self.ref_pair_style

        if len(self.ref_pair_style) == 1:
            return self.ref_pair_style[0]
        else:  # Here the tricky part. I need to create hybrid overlay ...
            full_pair_style = "hybrid/overlay "
            for ps in self.ref_pair_style:
                full_pair_style += f"{ps} "
            return full_pair_style

# ========================================================================== #
    def get_ref_pair_coeff(self):
        """
        Return the pair_coeff for the reference calculations
        """
        self._set_directories()
        if len(self.ref_pair_style) == 1:
            return self.ref_pair_coeff
        else:
            ref_pair_coeff = []
            for ps, pc in zip(self.ref_pair_style, self.ref_pair_coeff):
                refpssplit = ps.split()
                for ppc in pc:
                    refpcsplit = ppc.split()
                    refpc = " ".join([*refpcsplit[:2],
                                      refpssplit[0],
                                      *refpcsplit[2:]])
                    ref_pair_coeff.append(refpc)
            return ref_pair_coeff

# ========================================================================== #
    def _get_pair_style(self):
        self._set_directories()
        # We need to create the hybrid/overlay format of LAMMPS
        if not isinstance(self.ref_pair_style, list):
            self.ref_pair_style = [self.ref_pair_style]

        if len(self.ref_pair_style) == 1:
            full_pair_style = f"hybrid/overlay {self.ref_pair_style[0]} " + \
                              f"{self.model.pair_style}"
        else:
            full_pair_style = "hybrid/overlay "
            for ps in self.ref_pair_style:
                full_pair_style += f"{ps} "
            full_pair_style += f"{self.model.pair_style}"

        return full_pair_style

# ========================================================================== #
    def _get_pair_coeff(self):
        self._set_directories()
        if not isinstance(self.ref_pair_style, list):
            self.ref_pair_style = [self.ref_pair_style]

        # First let's take care of only one reference potential
        if len(self.ref_pair_style) == 1:
            refpssplit = self.ref_pair_style[0].split()
            full_pair_coeff = []
            for refpc in self.ref_pair_coeff:
                refpcsplit = refpc.split()
                full_pair_coeff.append(" ".join([*refpcsplit[:2],
                                       refpssplit[0],
                                       *refpcsplit[2:]]))
            mlpcsplit = self.model.pair_coeff[0].split()
            mlpssplit = self.model.pair_style.split()
            mlpc = " ".join([*mlpcsplit[:2],
                             mlpssplit[0],
                             *mlpcsplit[2:]])
            full_pair_coeff.append(mlpc)

        # And now with an overlay reference potential
        else:
            full_pair_coeff = []
            for ps, pc in zip(self.ref_pair_style, self.ref_pair_coeff):
                refpssplit = ps.split()
                for ppc in pc:
                    refpcsplit = ppc.split()
                    refpc = " ".join([*refpcsplit[:2],
                                      refpssplit[0],
                                      *refpcsplit[2:]])
                    full_pair_coeff.append(refpc)
            mlpcsplit = self.model.pair_coeff[0].split()
            mlpssplit = self.model.pair_style.split()
            mlpc = " ".join([*mlpcsplit[:2],
                             mlpssplit[0],
                             *mlpcsplit[2:]])
            full_pair_coeff.append(mlpc)
        return full_pair_coeff

# ========================================================================== #
    def predict(self, atoms, coef=None):
        """
        Function that gives the mlip_energy
        """
        return self.model.predict(atoms, coef)

# ========================================================================== #
    def update_matrices(self, atoms):
        """
        """

        # First compute reference energy/forces/stress
        if isinstance(atoms, Atoms):
            atoms = [atoms]

        calc = LAMMPS(pair_style=self.get_ref_pair_style(lmp=True),
                      pair_coeff=self.get_ref_pair_coeff(),
                      atom_style=self.ref_atom_style)

        if self.model_post is not None:
            calc.set(model_post=self.ref_model_post)

        dummy_at = []
        for at in atoms:
            at0 = at.copy()
            at0.calc = calc
            refe = at0.get_potential_energy()
            reff = at0.get_forces()
            refs = at0.get_stress()

            dumdum = at.copy()

            e = at.get_potential_energy() - refe
            f = at.get_forces() - reff
            s = at.get_stress() - refs
            spcalc = SinglePointCalculator(dumdum,
                                           energy=e,
                                           forces=f,
                                           stress=s)
            dumdum.calc = spcalc
            dummy_at.append(dumdum)

        # Now get descriptor features
        self.model.update_matrices(dummy_at)
        self.nconfs = self.model.nconfs

# ========================================================================== #
    def next_coefs(self, mlip_coef, *args, **kwargs):
        """
        """
        msg = self.model.next_coefs(mlip_coef, *args, **kwargs)
        return msg

# ========================================================================== #
    def train_mlip(self):
        """
        """
        self._set_directories()
        self.model.train_mlip()

    # GA: Need to overwrite this abstract methods, but I'm not sure
    #     if it is used at all.
    def get_mlip_energy(coef, desc):
        """
        Function that gives the mlip_energy
        """
        raise NotImplementedError

# ========================================================================== #
    def get_calculator(self):
        """
        Initialize a ASE calculator from the model
        """
        calc = LAMMPS(pair_style=self.pair_style,
                      pair_coeff=self.pair_coeff,
                      atom_style=self.atom_style,
                      keep_alive=False)
        if self.model_post is not None:
            calc.set(model_post=self.model_post)
        return calc

# ========================================================================== #
    def __str__(self):
        txt = "Delta Learning potential\n"
        txt += f"Reference pair_style: {self.ref_pair_style}\n"
        txt += f"Reference pair_coeff: {self.ref_pair_coeff}\n"
        txt += str(self.model)
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = "Delta learning potential\n"
        txt += "------------------------\n"
        txt += "Reference potential :\n"
        txt += f"pair_style {self.ref_pair_style}\n"
        for pc in self.ref_pair_coeff:
            txt += f"pair_coeff {pc}\n\n"
        txt += "MLIP potential :\n"
        txt += repr(self.model)
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np


try:
    import pandas as pd
except ImportError:
    pass

def make_dataframe(df, name, atoms, atomic_env,
                   energy=None, forces=None, we=None, wf=None):
    """
    Append atoms information to the dataframe.
    Return the dataframe WITHOUT writing in a file

    2 modes : 1. We don't yet know energy, forces, we, wf
              2. We also add energy, forces, we, wf

    name: :class:`list` :shape:`[nconfs]`
        Name of each configuration

    atoms: :class:`list` :shape:`[nconfs]`
        Ase.Atoms object for each configuration

    atomic_env: :class:`list` :shape:`[nconfs]`
        pyace.catomicenvironment.ACEAtomicEnvironment for each confs

    energy: :class:`np.array` :shape:`[nconfs]`
        Energy of each configuration

    forces: :class:`list` :shape:`[nconfs][natoms:3]`
        list of np.array containing the forces on each atoms

    we: :class:`np.array` :shape:`[nconfs]`
        Weight of each configuration

    wf: :class:`list` :shape:`[nconfs][natoms]`
        list of np.array containing the weight for the forces on each atom
        The sum must be equal to 1.
        The relative weight between e and f is given by alpha
    """
    add_ef = all(_ is not None for _ in (energy, forces, we, wf))
    nat = np.array([])
    for at in atoms:
        nat = np.append(nat, len(at))

    # Creating the dict and adding it to the df
    if add_ef:
        to_add = dict(name=name, ase_atoms=atoms,
                      energy_corrected=energy, forces=forces,
                      NUMBER_OF_ATOMS=nat, atomic_env=atomic_env,
                      w_energy=we, w_forces=wf)
    else:
        to_add = dict(name=name, ase_atoms=atoms,
                      NUMBER_OF_ATOMS=nat, atomic_env=atomic_env)

    return pd.DataFrame(to_add)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from pathlib import Path

from ase.io.abinit import read_abinit_out
from ase.calculators.singlepoint import SinglePointCalculator as SPCalc
from mlacs.utilities import AbinitNC

from ... import context  # noqa


@pytest.fixture()
def build_ncobj():
    root = Path()
    files = []
    for f in root.rglob('*nc'):
        _nc = AbinitNC()
        _nc.ncfile = f
    yield files


@pytest.fixture
@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_atoms_from_ncfiles(root, build_ncobj):
    """
    """
    f = root / 'reference_files'
    with open(f / 'abinit.abo', 'r') as fd:
        results = read_abinit_out(fd)

    atoms = results.pop("atoms")
    energy = results.pop("energy")
    forces = results.pop("forces")
    stress = results.pop("stress")

    calc = SPCalc(atoms,
                  energy=energy,
                  forces=forces,
                  stress=stress)
    calc.version = results.pop("version")
    atoms.calc = calc

    for ncobj in build_ncobj:
        ncobj.read()
        ncatoms = ncobj.convert_to_atoms()
        assert atoms == ncatoms


@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_dict_from_ncfile(root, build_ncobj):
    """
    """

    fulldict = dict()
    _nc = AbinitNC()
    for f in root.rglob('*nc'):
        fulldict.update(_nc.read(f))

    adddict = dict()
    for ncobj in build_ncobj:
        adddict.update(ncobj.read())

    for key, val in adddict.items():
        assert (fulldict[key] == val).all()



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .lammps_state import BaseLammpsState
from ..utilities.io_lammps import (LammpsBlockInput,
                                   EmptyLammpsBlockInput)


# ========================================================================== #
# ========================================================================== #
class OptimizeLammpsState(BaseLammpsState):
    """
    Class to manage geometry optimizations with LAMMPS.

    Parameters
    ----------
    min_style: :class:`str`
        Choose a minimization algorithm to use when a minimize command is
        performed.
        The options are the one available with the ``min_style`` command
        of LAMMPS.

        - `cg`
        - `hftn`
        - `sd`
        - `quickmin`
        - `fire`

        Default `cg`.

    etol: :class:`float`
        Stopping tolerance for energy
        Default ``0.0``

    ftol: :class:`float`
        Stopping tolerance for forces
        Default ``1.0e-6``

    dt : :class:`float` (optional)
        Timestep, in fs. Default ``0.5`` fs.

    pressure: :class:`float` or ``None`` (optional)
        Target pressure for the optimization, in GPa.
        Only available if min_style is 'cg'.
        If ``None``, no cell relaxation is applied.
        Default ``None``

    ptype: ``iso`` or ``aniso`` (optional)
        Only available if min_style is 'cg'.
        Handle the type of pressure applied. Default ``iso``

    vmax: ``iso`` or ``aniso`` (optional)
        The vmax keyword can be used to limit the fractional change in the
        volume of the simulation box that can occur in one iteration of
        the minimizer.
        Default ``1.0e-3``

    nsteps : :class:`int` (optional)
        Maximum number of minimizer iterations during production phase.
        Also sets up the max number of force/energy evaluations.
        Default ``10000`` steps.

    nsteps_eq : :class:`int` (optional)
        Maximum number of minimizer iterations during equilibration phase.
        Also sets up the max number of force/energy evaluations.
        Default ``1000`` steps.

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    blocks : :class:`LammpsBlockInput` or :class:`list` (optional)
        Custom block input class. Can be a list of blocks.
        If ``None``, nothing is added in the input. Default ``None``.

    Examples
    --------

    >>> from ase.io import read
    >>> initial = read('A.traj')
    >>>
    >>> from mlacs.state import OptimizeLammpsState
    >>> neb = OptimizeLammpsState(initial, pressure=0, ptype='iso')
    >>> state.run_dynamics(initial, mlip.pair_style, mlip.pair_coeff)
    """

    def __init__(self, min_style="cg", etol=0.0, ftol=1e-6, dt=0.5,
                 pressure=None, ptype="iso", vmax=1e-3,
                 nsteps=1000, nsteps_eq=100, logfile=None, trajfile=None,
                 loginterval=50, blocks=None, **kwargs):
        super().__init__(nsteps, nsteps_eq, logfile, trajfile, loginterval,
                         blocks, **kwargs)

        self.min_style = min_style
        self.dt = dt
        self.pressure = pressure
        self.ptype = ptype
        self.vmax = vmax

        self.criterions = (etol, ftol)

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        return EmptyLammpsBlockInput("empty_thermostat")

# ========================================================================== #
    def _get_block_run(self, eq):
        etol, ftol = self.criterions
        if eq:
            nsteps = self.nsteps_eq
        else:
            nsteps = self.nsteps

        block = LammpsBlockInput("optimization", "Geometry optimization")

        if self.pressure is not None:
            txt = f"fix box all box/relax {self.ptype} " + \
                  f"{self.pressure*10000} vmax {self.vmax}"
            block("press", txt)
        block("thermo", "thermo 1")
        block("min_style", f"min_style {self.min_style}")
        block("minimize", f"minimize {etol} {ftol} {nsteps} {2*nsteps}")
        return block

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "Geometry optimization as implemented in LAMMPS\n"
        if self.pressure is not None:
            msg += f"   target pressure: {self.pressure}\n"
        msg += f"   min_style: {self.min_style}\n"
        msg += f"   energy tolerance: {self.criterions[0]}\n"
        msg += f"   forces tolerance: {self.criterions[1]}\n"
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ase.units import Bohr, Hartree, GPa, bar


# ========================================================================== #
def unit_converter(obs_arr,
                   obs_unit,
                   target='ASE'):
    """
    Convert observable from atomic units to target units specified by `target`.

    Keywords of supported `target_style` systems of units:
        - custom units -> `custom`
        - ASE units (see ase.units) -> `ASE`
        - LAMMPS `metal` units -> `metal`

    Parameters
    ----------

    obs_arr: :class:`numpy.ndarray`
        Array of observable in atomic units.

    obs_unit: :class:`str`
        Unit of `obs_arr`.

    target: :class:`str` (optional)
        String defining the target unit system.
        Default: `ASE`.

    Returns
    ----------

    new_arr: :class:`numpy.ndarray`
        Converted observable.

    new_unit: :class:`str`
        Unit of `new_arr`.

    Notes
    ----------

    `custom` units
        - energy: 'eV',
        - distance: 'Ang',
        - volume: 'Ang^3',
        - force: 'eV/Ang',
        - pressure/stress: 'GPa'

    `metal` units (cf. LAMMPS)
        - energy: 'eV',
        - distance: 'Ang',
        - volume: 'Ang^3',
        - force: 'eV/Ang',
        - pressure/stress: 'bar'

    `ASE` units
        - energy: 'eV',
        - distance: 'Ang',
        - volume: 'Ang^3',
        - force: 'eV/Ang',
        - pressure/stress: 'eV/Ang^3'
    """
    Ha2eV = Hartree
    Bohr2Ang = Bohr

    # Store all data on conversions
    stl_dict = {}
    stl_dict['custom'] = {'Ha': [Ha2eV, 'eV'],
                          'Bohr': [Bohr2Ang, 'Ang'],
                          'Ha/Bohr': [Ha2eV/Bohr2Ang, 'eV/Ang'],
                          'Bohr^3': [Bohr2Ang**3, 'Ang^3'],
                          'Ha/Bohr^3': [(Ha2eV/Bohr2Ang**3)/GPa, 'GPa'],
                          }
    stl_dict['metal'] = {'Ha': [Ha2eV, 'eV'],
                         'Bohr': [Bohr2Ang, 'Ang'],
                         'Ha/Bohr': [Ha2eV/Bohr2Ang, 'eV/Ang'],
                         'Bohr^3': [Bohr2Ang**3, 'Ang^3'],
                         'Ha/Bohr^3': [(Ha2eV/Bohr2Ang**3)/bar, 'bar'],
                         }
    stl_dict['ASE'] = {'Ha': [Ha2eV, 'eV'],
                       'Bohr': [Bohr2Ang, 'Ang'],
                       'Ha/Bohr': [Ha2eV/Bohr2Ang, 'eV/Ang'],
                       'Bohr^3': [Bohr2Ang**3, 'Ang^3'],
                       'Ha/Bohr^3': [(Ha2eV/Bohr2Ang**3), 'eV/Ang^3'],
                       }
    # unit_convert_dict format: {'atomic_unit': [conv_factor, 'target_unit']}
    unit_convert_dict = stl_dict[target]

    # Conversion itself
    if obs_unit in unit_convert_dict:
        conv_factor, new_unit = unit_convert_dict[obs_unit]
        new_arr = obs_arr*conv_factor
        return new_arr, new_unit
    else:
        return obs_arr, obs_unit



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
try:
    from sklearn.gaussian_process import GaussianProcessRegressor
    from sklearn.gaussian_process.kernels import (RBF,
                                                  ConstantKernel as C,
                                                  WhiteKernel)
except ImportError:
    msg = "You need sklearn to use the calphagpy modules"
    raise ModuleNotFoundError(msg)


default_gp_parameters = {"n_restarts_optimizer": 100,
                         "normalize_y": True,
                         "alpha": 1e-10}


# ========================================================================== #
# ========================================================================== #
class GaussianProcessInterface:
    """
    Base class to interface gaussian process from scikit-learn

    Parameters
    ----------

    ndim: :class:`int`
        The number of dimensions for the feature
    kernel: :class:`Kernel`
        The kernel of the gaussian process.
        Default are :
        RBF() * C() + WhiteKernel()
    gp_parameters: :class:`dict`
        Parameters for the GaussianProcessRegressor object.
        The default parameters are :
        {"n_restart_optimizer": 100, "normalyze_y": True, "alpha": 1e-10}
    """
    def __init__(self,
                 ndim=1,
                 kernel=RBF()*C()+WhiteKernel,
                 gp_parameters={}):

        default_gp_parameters.update(gp_parameters)
        self.gp = GaussianProcessRegressor(kernel=kernel,
                                           **default_gp_parameters)

        self.trained = False

        self.ndim = ndim
        self.x = None
        self.y = None

# ========================================================================== #
    def add_new_data(self, x, y):
        """
        Add new data to train the gaussian process

        Parameters
        ----------

        x: :class:`np.ndarray`
            The features, with dimension (nsamples, nfeatures)
        y: :class:`np.ndarray`
            The target values, with dimension (nsamples, ndim)
        """
        self._add_new_data(x, y)

# ========================================================================== #
    def _add_new_data(self, x, y):
        """
        """
        # We need to ensure that x and y has the right dimensions
        if len(x.shape) == 1 and self.ndim == 1:
            x = x.reshape(-1, 1)
        elif len(x.shape) == 1 and self.ndim > 1:
            x = x.reshape(1, -1)

        if self.x is None:
            self.x = x
            self.y = y
        else:
            self.x = np.r_[self.x, x]
            self.y = np.r_[self.y, y]

# ========================================================================== #
    def train(self):
        """
        Train the gaussian process.
        """
        if self.x is None or self.y is None:
            msg = "You need to add data with the add_new_data() function " + \
                  "to train the gaussian process"
            raise RuntimeError(msg)

        # Do the fit
        self.gp.fit(self.x, self.y)
        self.trained = True

        # Get some lower bounds for the predictions
        self.lb = self.x.min(axis=0)
        self.ub = self.x.max(axis=0)

# ========================================================================== #
    def predict(self, x, return_cov=False):
        """
        Predict values for a given x

        Parameters
        ----------

        x: :class:`np.ndarray`
            The features, with dimension (nsamples, nfeatures)
        """
        if not self.trained:
            msg = "You need to train the gaussian process before doing " + \
                  "predictions"
            raise RuntimeError(msg)

        # We need to have everything as array so that sklearn is happy
        if isinstance(x, (float, int)):
            x = np.array([x])
        elif isinstance(x, list):
            x = np.array(x)

        # We need to ensure that x has the right dimensions
        if len(x.shape) == 1 and self.ndim == 1:
            x = x.reshape(-1, 1)
        elif len(x.shape) == 1 and self.ndim > 1:
            x = x.reshape(1, -1)
        if return_cov:
            y, y_cov = self.gp.predict(x, return_cov=True)
            return y, y_cov
        else:
            y = self.gp.predict(x)
            return y



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.build import bulk

from ... import context  # noqa
from mlacs.mlip.utilities import split_dataset


def test_split_dataset():
    """

    """
    rng = np.random.default_rng()
    # First we create fake dataset
    allconfs = []
    for i in range(10):
        at = bulk("Si", cubic=True)
        at.rattle(rng=rng)
        at.info = dict(number=i)
        allconfs.append(at)

    trainset, testset = split_dataset(allconfs, rng=rng)
    # First let's check that we have the right number of configurations
    assert len(trainset) == 5
    assert len(testset) == 5

    #  Now let's check that configurations are differents
    train_idx = np.array([at.info["number"] for at in trainset])
    test_idx = np.array([at.info["number"] for at in testset])
    for i in train_idx:
        assert i not in test_idx

    # Let's set up a different train ratio
    trainset, testset = split_dataset(allconfs, 0.3, rng=rng)
    # First let's check that we have the right number of configurations
    assert len(trainset) == 3
    assert len(testset) == 7

    #  Now let's check that configurations are differents
    train_idx = np.array([at.info["number"] for at in trainset])
    test_idx = np.array([at.info["number"] for at in testset])
    for i in train_idx:
        assert i not in test_idx



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
import matplotlib.pyplot as plt
from ..utilities.plots import plot_error, init_rcParams


def main(args, parser):

    data = np.loadtxt(args.file)
    if args.datatype not in ["energy", "forces", "stress", None]:
        raise ValueError("The type argument has to be "
                         "energy, forces or stress")
    rmse = True
    if args.normse:
        rmse = False
    mae = True
    if args.nomae:
        mae = False
    rsquared = True
    if args.nor2:
        rsquared = False
    figsize = (float(args.figsize), float(args.figsize))
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    init_rcParams()
    ax = fig.add_subplot()
    plot_error(ax,
               data,
               datatype=args.datatype,
               showrmse=rmse,
               showmae=mae,
               showrsquared=rsquared)
    if args.save is not None:
        plt.savefig(args.save)
    if not args.noshow:
        plt.show()


class CLICommand:
    """Plot the distribution of error for the MLIP


    Example:

        $ mlacs error MLIP-Energy_comparison.dat -datatype energy
    """
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('file', help="file with the data")
        parser.add_argument("-s", "--save", default=None,
                            help="Name of the file to save the plot. "
                                 "Default None")
        parser.add_argument("--noshow", action="store_true",
                            help="To disable the visualisation of the plot")
        parser.add_argument('--datatype', default=None,
                            help="Type of data in the file. Can be "
                            "energy, forces or stress")
        parser.add_argument('--nomae', action="store_true",
                            help="To remove the MAE from the plot")
        parser.add_argument('--normse', action="store_true",
                            help="To remove rmse from the plot")
        parser.add_argument('--nor2', action="store_true",
                            help="to remove the r^2 from the plot")
        parser.add_argument("--figsize", default="10",
                            help="Size of the figure for matplotlib")

    @staticmethod
    def run(args, parser):
        main(args, parser)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np

from ase.atoms import Atoms
from ase.io import write
from ase.io.lammpsdata import write_lammps_data
from ase.calculators.singlepoint import SinglePointCalculator as SPC

from .lammps_state import BaseLammpsState
from ..core import PathAtoms
from ..core.manager import Manager
from ..utilities.io_lammps import (LammpsBlockInput,
                                   EmptyLammpsBlockInput,
                                   get_lammps_command)


# ========================================================================== #
# ========================================================================== #
class NebLammpsState(BaseLammpsState):
    """
    Class to manage Nudged Elastic Band (NEB) calculation with LAMMPS.
    This class is a part of TransPath objects, meaning that it produces
    positions interpolation according to a reaction coordinate.

    Parameters
    ----------
    images: :class:`list` or `PathAtoms`
        mlacs.PathAtoms or list of ase.Atoms object.
        The list contain initial and final configurations of the reaction path.

    xi: :class:`numpy.array` or `float`
        Value of the reaction coordinate for the constrained MD.
        Default ``None``

    min_style: :class:`str`
        Choose a minimization algorithm to use when a minimize command is
        performed. Default `quickmin`.

    Kspring: :class:`float`
        Spring constante for the NEB calculation.
        Default ``1.0``

    etol: :class:`float`
        Stopping tolerance for energy
        Default ``0.0``

    ftol: :class:`float`
        Stopping tolerance for energy
        Default ``1.0e-3``

    dt : :class:`float` (optional)
        Timestep, in fs. Default ``1.5`` fs.

    nimages : :class:`int` (optional)
        Number of images used along the reaction coordinate. Default ``1``.
        which is suposed the saddle point.

    nprocs : :class:`int` (optional)
        Total number of process used to run LAMMPS.
        Have to be a multiple of the number of images.
        If nprocs > than nimages, each image will be parallelized using the
        partition scheme of LAMMPS.
        Per default it assumes that nprocs = nimages

    mode: :class:`float` or :class:`string`
        Value of the reaction coordinate or sampling mode:
        - ``float`` sampling at a precise coordinate.
        - ``rdm_true`` randomly return the coordinate of an images.
        - ``rdm_spl`` randomly return the coordinate of a splined images.
        - ``rdm_memory`` homogeneously sample the splined reaction coordinate.
        - ``None`` return the saddle point.
        Default ``saddle``

    linear : :class:`Bool` (optional)
        If true, the reaction coordinate is a linear interpolation.
        Default ``False``

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    prt : :class:`Bool` (optional)
        Printing options. Default ``True``

    Examples
    --------

    >>> from ase.io import read
    >>> initial = read('A.traj')
    >>> final = read('B.traj')
    >>>
    >>> from mlacs.state import NebLammpsState
    >>> neb = NebLammpsState([initial, final])
    >>> state.run_dynamics(None, mlip.pair_style, mlip.pair_coeff)
    """

    def __init__(self, images, xi=None,
                 min_style="quickmin", Kspring=1.0, etol=0.0, ftol=1.0e-3,
                 dt=1.5, nimages=4, nprocs=None, mode=None, interval=None,
                 linear=False, print=False,
                 nsteps=1000, nsteps_eq=100, logfile=None, trajfile=None,
                 loginterval=50, blocks=None, **kwargs):

        super().__init__(nsteps, nsteps_eq, logfile, trajfile, loginterval,
                         blocks, **kwargs)

        self.dt = dt
        self.pressure = None

        self._step = 1
        self.style = min_style
        self.criterions = (etol, ftol)
        self.nprocs = nprocs
        self.nreplica = nimages
        self.atomsfname = "atoms-0.data"
        self.print = print
        self.Kspring = Kspring
        self.patoms = images
        if not isinstance(self.patoms, PathAtoms):
            self.patoms = PathAtoms(self.patoms, interval=interval)
        if xi is not None:
            self.patoms.xi = xi
        if mode is not None:
            self.patoms.mode = mode

        self.linear = linear

# ========================================================================== #
    @Manager.exec_from_path
    def _write_lammps_atoms(self, atoms, atom_style, elements=None):
        """

        """
        write_lammps_data(self.atomsfname,
                          self.patoms.initial,
                          velocities=False,
                          atom_style=atom_style)
        instr = '# Final coordinates of the NEB calculation.\n'
        instr += '{0}\n'.format(len(self.patoms.final))
        for atoms in self.patoms.final:
            instr += '{} {} {} {}\n'.format(atoms.index+1, *atoms.position)
        with open("atoms-1.data", "w") as w:
            w.write(instr)

# ========================================================================== #
    def _get_block_init(self, atom_style, pbc, el, masses):
        """

        """
        pbc = "{0} {1} {2}".format(*tuple("sp"[int(x)] for x in pbc))

        block = LammpsBlockInput("init", "Initialization")
        block("units", "units metal")
        block("boundary", f"boundary {pbc}")
        block("atom_style", f"atom_style {atom_style}")
        block("atom_modify", "atom_modify  map array sort 0 0.0")
        txt = "neigh_modify every 2 delay 10" + \
              " check yes page 1000000 one 100000"
        block("neigh_modify", txt)
        block("read_data", "read_data atoms-0.data")
        for i, mass in enumerate(masses):
            block(f"mass{i}", f"mass {i+1}  {mass}")
        return block

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        return EmptyLammpsBlockInput("empty_thermostat")

# ========================================================================== #
    def _get_block_lastdump(self, el, eq):
        return EmptyLammpsBlockInput("empty_lastdump")

# ========================================================================== #
    def _get_atoms_results(self, initial_charges):
        """

        """
        self.patoms.images = self.extract_NEB_configurations()
        self.patoms.update
        atoms = self.patoms.splined
        if initial_charges is not None:
            atoms.set_initial_charges(initial_charges)
        # RB: Usefull for tests, we should move these in the Mlminimizer.
        i = self._step
        if self.print:
            if self.patoms._gpi is not None:
                if 4 < len(self.patoms._gpi.y):
                    import numpy as np
                    x = np.linspace(0, 1, 1001)
                    m, s = self.patoms._gpi.predict(x, True)
                    err = self.patoms._gp_error
                    np.savetxt(str(self.subsubdir / f'mean_{i:02d}.dat'), m)
                    np.savetxt(str(self.subsubdir / f'err_{i:02d}.dat'), err)
                    np.savetxt(str(self.subsubdir / f'max_{i:02d}.dat'),
                               np.max(s, axis=0))
                    np.savetxt(str(self.subsubdir / f'min_{i:02d}.dat'),
                               np.min(s, axis=0))
            write(str(self.subsubdir / f'pos_neb_images_{i:02d}.xyz'),
                  self.patoms.images, format='extxyz')
            write(str(self.subsubdir / f'pos_neb_splined_{i:02d}.xyz'),
                  self.patoms.splined, format='extxyz')
        self._step += 1
        return atoms

# ========================================================================== #
    def _get_block_run(self, eq):
        etol, ftol = self.criterions

        block = LammpsBlockInput("transpath", "Transition Path")
        block("thermo", "thermo 1")
        block("timestep", f"timestep {self.dt / 1000}")
        block("fix_neb", f"fix neb all neb {self.Kspring} parallel ideal")
        block("run", "run 100")
        block("reset", "reset_timestep 0")
        block("image", "variable i equal part")
        block("min_style", f"min_style {self.style}")
        if self.linear:
            block("neb", f"neb {etol} {ftol} 1 1 1 final atoms-1.data")
        else:
            block("neb", f"neb {etol} {ftol} 200 100 1 final atoms-1.data")
        block("write_data", "write_data neb.$i")
        return block

# ========================================================================== #
    def extract_NEB_configurations(self):
        """
        Extract the positions and energies of a NEB calculation for all
        replicas.
        """
        img_at = []

        def set_atoms(C, R):
            Z = self.patoms.initial.get_atomic_numbers()
            at = Atoms(numbers=Z, positions=R, cell=C)
            at.set_pbc(True)
            return at

        for rep in range(int(self.nreplica)):
            nebfile = str(self.path / f'neb.{rep}')
            positions, cell = self._read_lammpsdata(nebfile)
            check = False
            with open(self.path / f'log.lammps.{rep}') as r:
                for _ in r:
                    if check:
                        etotal = _.split()[2]
                        break
                    if 'initial, next-to-last, final =' in _:
                        check = True
            atoms = set_atoms(cell, positions)
            calc = SPC(atoms=atoms, energy=etotal)
            atoms.calc = calc
            img_at.append(atoms)
        return img_at

# ========================================================================== #
    def _get_lammps_command(self):
        '''
        Function to load the batch command to run LAMMPS with replica.
        '''
        cmd = get_lammps_command()
        exe = cmd.split()[-1]

        if "-partition" in cmd:
            _ = cmd.split().index('-n')+1
            self.nprocs = int(cmd.split()[_])
            self.nreplica = int(cmd.split('x')[0][-1])
            return f"{cmd} -in {self.lammpsfname} -sc out.lmp"

        if self.nreplica is not None and self.nprocs is not None:
            pass
        elif self.nreplica is not None and self.nprocs is None:
            if '-n' in cmd:
                _ = cmd.split().index('-n')+1
                self.nprocs = int(cmd.split()[_])
            else:
                self.nprocs = self.nreplica
        elif self.nreplica is None and self.nprocs is not None:
            self.nreplica = self.nprocs
        else:
            if '-n' in cmd:
                _ = cmd.split().index('-n')+1
                self.nprocs = int(cmd.split()[_])
                self.nreplica = self.nprocs
            else:
                self.nreplica, self.nprocs = 1, 1

        n1, n2 = self.nreplica, self.nprocs // self.nreplica
        if n2 == 0:
            n2 = 1
        cmd = f"mpirun -n {int(n1*n2)} {exe} -partition {n1}x{n2}"
        return f"{cmd} -in {self.lammpsfname} -sc out.lmp"

# ========================================================================== #
    def _read_lammpsdata(self, filename, wrap=True):
        """
        Extract positions from lammpsdata files with memory of periodicity.
        Inspired from ASE.
        """
        (xy, xz, yz) = None, None, None
        (section, style) = None, None
        pos_in = {}
        travel_in = {}

        with open(filename, 'r') as r:
            for _ in r:
                if 'atoms' in _:
                    N = int(_.split()[0])
                if 'Atoms' in _:
                    (section, _, style) = _.split()
                    continue
                if 'Velocities' in _:
                    (section) = _.split()
                    continue
                if 'xlo xhi' in _:
                    (xlo, xhi) = [float(x) for x in _.split()[0:2]]
                if 'ylo yhi' in _:
                    (ylo, yhi) = [float(x) for x in _.split()[0:2]]
                if 'zlo zhi' in _:
                    (zlo, zhi) = [float(x) for x in _.split()[0:2]]
                if 'xy xz yz' in _:
                    (xy, xz, yz) = [float(x) for x in _.split()[0:3]]
                if section == 'Atoms':
                    fields = _.split()
                    lenght = len(fields)
                    if lenght == 0:
                        continue
                    id = int(fields[0])
                    if style == "atomic" and (lenght == 5 or lenght == 8):
                        # id type x y z [tx ty tz]
                        pos_in[id] = (
                            int(fields[1]),
                            float(fields[2]),
                            float(fields[3]),
                            float(fields[4]),
                        )
                        if lenght == 8:
                            travel_in[id] = (
                                int(fields[5]),
                                int(fields[6]),
                                int(fields[7]),
                            )
                    else:
                        msg = f"Style '{style}' not supported or" + \
                              f"invalid number of fields {lenght}"
                        raise RuntimeError(msg)

        # set cell
        cell = np.zeros((3, 3))
        cell[0, 0] = xhi - xlo
        cell[1, 1] = yhi - ylo
        cell[2, 2] = zhi - zlo
        if xy is not None:
            cell[1, 0] = xy
        if xz is not None:
            cell[2, 0] = xz
        if yz is not None:
            cell[2, 1] = yz
        positions = np.zeros((N, 3))
        for id in pos_in.keys():
            ind = id - 1
            positions[ind, :] = [pos_in[id][1]+cell[0, 0]*travel_in[id][0],
                                 pos_in[id][2]+cell[1, 1]*travel_in[id][1],
                                 pos_in[id][3]+cell[2, 2]*travel_in[id][2]]
        return positions, cell

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "NEB calculation as implemented in LAMMPS\n"
        msg += f"Number of replicas :                     {self.nreplica}\n"
        msg += f"String constant :                        {self.Kspring}\n"
        msg += f"Sampling mode :                          {self.patoms.mode}\n"
        msg += "\n"
        return msg
